#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (C) 2023-2025 Pauline Hamon-Giraud and Alice Mataigne - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

#import warnings
import os
import json
import numpy as np
import pandas as pd
from ete3 import NCBITaxa
import plotly.express as px
import plotly.graph_objects as go
from plotly.io import write_json
from plotly.subplots import make_subplots

from ontosunburst.ontosunburst import ontosunburst
from esmecata.report.esmecata_compression import RANK2COL, RANK_SORTED

# from statistics import NormalDist
# import matplotlib.pyplot as plt
# import seaborn as sns
# import plotly.tools as tls
# import math
# import scipy.stats as st


__author__ = "Pauline Hamon-Giraud, Alice Mataigne"

# warnings.filterwarnings('ignore')

# ================
# Global variables
# ================

PLOT_BGCOLOR='#e6ffe6'
LINECOLOR = 'DarkSlateGrey'
COLORSCALE = px.colors.qualitative.Dark24
EC_CLASSES_COLORS = {"Ligases": "#636EFA",
    "Oxidoreductases": "#EF553B",
    "Transferases": "#00CC96",
    "Hydrolases": "#AB63FA",
    "Lyases": "#FFA15A",
    "Isomerases": "#19D3F3",
    "Translocases": "#FF6692"} # 'Plotly' default color map
DOWNLOAD_FMT = "svg"
FONTPARAMS = dict(
    family="Courier New, monospace",
    size=14,
    color="black"
)

# ========================
# Code factoring functions
# ========================

def _format_trace(fig):
    '''
    Set visual harmonization of a figure trace
    '''
    fig.update_traces(marker=dict(line=dict(width=1, color=LINECOLOR)))

def _format_axes(fig):
    '''
    Set visual harmonization of a figure axes

    Parameters:
        fig : a plotly object
    '''  
    fig.update_xaxes(
        showline=True, 
        linewidth=2, 
        linecolor=LINECOLOR, 
        mirror=True, 
        showgrid=True, 
        gridwidth=2)
    fig.update_yaxes(
        showline=True, 
        linewidth=2, 
        linecolor=LINECOLOR, 
        mirror=True, 
        showgrid=True, 
        gridwidth=2)
    
# def _create_config(format, filename):
#     config = {'remove': ['select', 'zoomIn', 'zoomOut', 'autoScale', 'lasso2d'],
#               'toImageButtonOptions': {'format': 'svg', # one of png, svg, jpeg, webp
#                                        'filename': 'custom_image'
#                 }
#             }
#     return config
    
# =============================
# Data gathering and formatting
# =============================

def _format_taxo(proteome_tax_id):
    '''
    Recovers the full taxonomic lineage of all OTUs/ASVs analyzed by esMeCaTa by parsing proteome_tax_id.tsv with ete3

        Parameters:
            proteome_tax_id (pandas): a DataFrame loaded with the output file proteome_tax_id.tsv of EsMeCaTa proteomes

        Returns:
            df (pandas): DataFrame with the full taxonomic lineage (one column = one rank)
    '''

    ncbi_ranks = ["superkingdom","kingdom","subkingdom","superphylum","phylum",
    "subphylum","infraphylum","superclass","class","subclass",
    "infraclass","cohort","subcohort","superorder","order",
    "suborder","infraorder","parvorder","superfamily","family",
    "subfamily","tribe","subtribe","genus","subgenus",
    "section","subsection","series","subseries","species group",
    "species subgroup","species","forma specialis","subspecies","varietas",
    "subvariety","forma","serogroup","serotype","strain","isolate"]

    ncbi = NCBITaxa()

    ranks = {}
    names = {}
    data = {}

    # ete3 to get lineages
    for index, row in proteome_tax_id.iterrows():
        tmp = ncbi.get_lineage(row['tax_id'])
        ranks[index] = ncbi.get_rank(tmp)
        names[index] = ncbi.get_taxid_translator(tmp)

    # rank:name correspondance
    for k in ranks.keys():
        data[k] = {}
        for k2,v in ranks[k].items():
            data[k].update({v:names[k][k2]})

    df = pd.DataFrame.from_dict(data, orient='index')
    if 'no rank' in df.columns:
        df.drop('no rank', axis=1, inplace=True)

    # order columns according to hierarchy
    ncbi_ranks = [rank for rank in ncbi_ranks if rank in df.columns]
    df = df[ncbi_ranks]

    return df

def post_analysis_config(input_table, results_path):
    '''
    Proceeds to format the inputs and outputs data in order to un all the others plotting functions efficiently

        Parameters:
            input_table (str): the path to the input table of taxa
            results_path (str): the path of the esmecata run

        Returns:
            data (dict) : gathers all useful data for the analysis of an esmecata run (inputs, summary tables, taxonomy ...)
    '''

    # Load and concat summaries of results
    stats_nb_proteomes = pd.read_table(
        os.path.join(results_path,'0_proteomes/stat_number_proteome.tsv'),
        sep='\t',
        header=0,
        index_col='observation_name')
    
    stats_nb_clustering = pd.read_table(
        os.path.join(results_path,'1_clustering/stat_number_clustering.tsv'),
        sep='\t',
        header=0,
        index_col='observation_name')
    
    stats_nb_annotation = pd.read_table(
        os.path.join(results_path,'2_annotation/stat_number_annotation.tsv'),
        sep='\t',
        header=0,
        index_col='observation_name')

    df_stats = stats_nb_proteomes.join([stats_nb_clustering, stats_nb_annotation])
    df_stats = df_stats.dropna()

    # Load input
    input_data = pd.read_table(
        input_table,
        sep='\t',
        header=0,
        index_col=None,
        usecols=['observation_name', 'taxonomic_affiliation'])
    
    # Number of input and output taxa (some are discarded along the pipeline if no results)
    n_out = stats_nb_annotation.shape[0]
    n_in = input_data.shape[0]

    # Load tax ids (outputs)
    proteome_tax_id = pd.read_csv(
        os.path.join(results_path, '0_proteomes/proteome_tax_id.tsv'),
        header=0,
        index_col='observation_name',
        sep='\t')

    proteome_tax_id = proteome_tax_id[proteome_tax_id.index.isin(df_stats.index)]
    proteome_tax_id['n_proteomes'] = proteome_tax_id['proteome'].str.split(pat=',').str.len()

    # Input taxonomy must be formatted ; ete3 is used to find the lineage
    output_taxo = _format_taxo(proteome_tax_id)
    df_stats = df_stats.join(output_taxo)

    input_data.set_index('observation_name', inplace=True)
    discarded = input_data[~input_data.index.isin(df_stats.index)]
    n_discarded = n_in - n_out

    # split discarded taxo str
    discarded = discarded['taxonomic_affiliation'].str.split(";", expand = True)

    # Association tax ids (inputs)
    with open(os.path.join(results_path, '0_proteomes/association_taxon_taxID.json')) as f:
        association_taxon_tax_id = json.load(f)

    data = {
        "INPUT_DATA": input_data,
        "DISCARDED": discarded, 
        "N_DISCARDED": n_discarded, 
        "DF_STATS": df_stats, 
        "N_IN": n_in, 
        "N_OUT": n_out, 
        "PROTEOME_TAX_ID": proteome_tax_id, 
        "ASSOCIATION_PROTEOME_TAX_ID": association_taxon_tax_id, 
    }

    return data

# ==============================
# Proteomes summary page figures
# ==============================

def distributions_by_ranks(df_stats, output_folder, n_bins=15, rank='phylum', savefig=True):
    '''
    For the given rank, plots stacked barplots of how many proteomes belongs to each taxa of this rank.

        Parameters:
            df_stats (pandas): a pandas df, output of post_analysis_config()
            output_folder (str) : path to output folder to save fig
            rank (str): taxonomic rank for color code. Must match the input. Default 'phylum'
            savefig (bool): boolean to save fig or not

        Returns:
            fig : a plotly html figure
    '''
    df_stats.sort_values(by=rank, ascending=False, inplace=True)

    fig = px.histogram(
        df_stats,
        x="Number_proteomes",
        color=rank,
        height=600,
        width=500,
        nbins=n_bins,
        color_discrete_sequence=COLORSCALE,
        title=f"Distribution of the number of proteomes found by {rank}")
    
    _format_trace(fig)
    _format_axes(fig)
    fig.update_layout(
        xaxis_title="Number of proteomes",
        yaxis_title="Number of taxa", 
        plot_bgcolor=PLOT_BGCOLOR,
        font=FONTPARAMS)
   
    if savefig:
        fig.write_image(os.path.join(output_folder, 'proteomes_distribution_by_ranks.pdf'))
        write_json(
            fig, 
            os.path.join(output_folder, 'proteomes_distribution_by_ranks.json'), 
            pretty=True)

    return fig

def n_prot_ec_go_correlations(df_stats, results_path, rank="phylum", savefig=True):
    '''
    Plots the correlation between the number of GO terms and the number of EC numbers. Each dot is a taxa. Dot sizes are linked to the number of proteomes found.

        Parameters:
            df_stats (pandas): a pandas df, output of post_analysis_config()
            results_path (str) : the path of the esmecata run
            rank (str): taxonomic rank for color code. Must match the input. Default 'phylum'

        Returns:
            fig : a plotly figure
    '''

    # TODO : add legend for markers size + colors ?
    sizes = df_stats['Number_proteomes']**0.75+5
    fig = px.scatter(
        data_frame=df_stats, 
        x="Number_ecs",
        y="Number_go_terms",
        size=sizes,
        hover_data=["Number_ecs", "Number_go_terms", "Number_proteomes"],
        color=rank,
        #color_discrete_sequence=COLORSCALE,
        width=800,
        height=600,
        labels={"Number_ecs": "Number of ECs", "Number_go_terms": "Number of GO terms", "Number_proteomes": "Number of proteomes"},
        title="Correlation between the number of EC, GO terms, and proteomes")
    
    _format_trace(fig)
    _format_axes(fig)
    fig.update_layout(
        plot_bgcolor=PLOT_BGCOLOR,
        font=FONTPARAMS)
    
    if savefig:
        fig.write_image(os.path.join(results_path, "3_analysis/proteomes_figures/correlation_ec_go_proteomes.pdf"))
        write_json(
            fig, 
            os.path.join(results_path, "3_analysis/proteomes_figures/correlation_ec_go_proteomes.json"), 
            pretty=True)

    return fig

def taxo_ranks_contribution(proteome_tax_id, results_path, n_bins=10, savefig=True):
    '''
    For the given rank, plots stacked barplots of how many proteomes belongs to each taxa of this rank.

        Parameters:
            proteome_tax_id (pandas): a pandas df, output of post_analysis_config()
            n_in (int): number of input taxa for esmecata
            n_out (int): number of output taxa by esmecata
            results_path (str) : the path of the esmecata run

        Returns:
            fig : a plotly figure
    '''

    fig = px.histogram(
        proteome_tax_id,
        x="n_proteomes",
        color="tax_rank",
        height=600,
        width=500,
        nbins=n_bins,
        # labels={"n_proteomes": "Number of proteomes"},
        title="Number of proteomes per taxonomic rank") 

    _format_trace(fig)
    _format_axes(fig)
    fig.update_layout(
        xaxis_title="Number of proteomes",
        yaxis_title="Number of inputs", 
        plot_bgcolor=PLOT_BGCOLOR,
        font=FONTPARAMS)

    if savefig:
        fig.write_image(os.path.join(results_path, "3_analysis/proteomes_figures/taxonomic_ranks_contribution.pdf"))
        write_json(
            fig, 
            os.path.join(results_path, "3_analysis/proteomes_figures/taxonomic_ranks_contribution.json"), 
            pretty=True)

    return fig

def compare_ranks_in_out(proteome_tax_id, association_taxon_tax_id, output_folder, savefig=True):
    '''
    Plots a heatmap displaying how many input taxonomic ranks were assigned to the same rank by esmecata and how many were assigned to higher tanks.

        Parameters:
            proteome_tax_id (pandas): a pandas df loaded with esmecata 'proteome_tax_id.tsv' file, output of post_analysis_config()
            association_taxon_tax_id (dict) : a dict of taxa loaded from json by post_analysis_config()
            output_folder (str) : path to output folder

        Returns:
            fig : a plotly figure
    '''
    rank_list = RANK_SORTED

    d_in = dict()
    ncbi = NCBITaxa()

    # Loop on each input (=each key of 0_proteomes/association_taxon_taxID.json)
    for taxa in association_taxon_tax_id.keys():
        # Get the list of rank names, ex: ['Gammaproteobacteria', 'Proteobacteria', 'Bacteria', 'cellular organisms']
        reversed_affiliation_taxa = list(reversed(list(association_taxon_tax_id[taxa].keys()))) # That line is f**** up

        i = 0
        found = False

        # Loop until the lowest known taxa rank is found
        # TODO : Discard taxa with only unknown
        while not found and i < len(reversed_affiliation_taxa):
            if "clade" not in reversed_affiliation_taxa[i] and association_taxon_tax_id[taxa][reversed_affiliation_taxa[i]] != ['not_found']:
                tax_id = association_taxon_tax_id[taxa][reversed_affiliation_taxa[i]]                
                tax_rank = ncbi.get_rank(tax_id) # get_rank takes a list for parameter (here, list of one elem, returns a list of one elem)
                # d_in[taxa] = out_asso_rank[list(tax_rank.values())[0]]
                d_in[taxa] = list(tax_rank.values())[0]
                found = True
            else:
                i += 1        
    
    d_out = dict()
    for ASV in proteome_tax_id.index:
        # rank = out_asso_rank[proteome_tax_id.loc[ASV, 'tax_rank']]
        rank = proteome_tax_id.loc[ASV, 'tax_rank']
        d_out[ASV] = rank

    matrix = pd.DataFrame(index=rank_list, columns=rank_list, dtype=int)
    matrix = matrix.fillna(int(0))    

    for ASV in d_out.keys():
        in_rank = d_in[ASV]
        out_rank = d_out[ASV]
        matrix.loc[in_rank, out_rank] += 1
    
    matrix = matrix.loc[(matrix != 0).any(axis=1)]
    matrix = matrix.loc[:,(matrix != 0).any(axis=0)]
    matrix = matrix.iloc[::-1] # Reverse rows for a more suited viz

    fig = px.imshow(
        matrix,
        text_auto=True,
        height=600,
        width=800,
        labels=dict(x="EsMeCaTa model rank", y="Taxa input rank", color="Number of proteomes"),
        title="Difference between input ranks and EsMeCaTa ranks",
        color_continuous_scale='YlGn')
    
    for i in range(0, matrix.shape[1]-1):
        fig.add_vline(x=i+0.5, line_width=2, line_color=LINECOLOR)
    for i in range(0, matrix.shape[0]-1):
        fig.add_hline(y=i+0.5, line_width=2, line_color=LINECOLOR)

    _format_axes(fig)
    fig.update_layout(font=FONTPARAMS)
    
    if savefig:
        fig.write_image(os.path.join(output_folder, 'input_and_output_ranks.pdf'))
        write_json(
            fig, 
            os.path.join(output_folder, 'input_and_output_ranks.json'), 
            pretty=True)
    
    return fig

# ============================================
# EC numbers and GO terms distribution figures
# ============================================

def create_annot_obs_df(dataset_annotation_file, outpath, content):
    ''' Prepare data of 'frequences of EC numbers / Go terms in taxa' and 'fraction of all EC numbers / Go terms in taxa'

    Args:
        dataset_annotation_file (str) : path to an EC*taxa or GO*taxa matrix, computed by create_dataset_annotation_file()
        outpath (str) : the path of the esmecata run
        content (str) : 'GO terms' or 'EC numbers' to specifies which data is processed

    Returns:
        data (dict) : stores two dataframes corresponding to each data, as well as summary numbers
    '''

    df = pd.read_csv(
        dataset_annotation_file,
        sep='\t',
        header=0,
        index_col=0)

    # Frequences of EC/GO terms in taxa
    # ---------------------------------

    # EC/GO*taxa matrix rather than taxa*EC/GO
    dfT = df.T

    # Get data
    dfT["frequency"] = dfT.ne(0).sum(axis=1)/dfT.shape[1]    
    dfT["annot_name"] = dfT.index

    if content == 'EC numbers':
            
        ec_class = {"1": "Oxidoreductases", "2": "Transferases", "3": "Hydrolases", 
                "4": "Lyases", "5": "Isomerases", "6": "Ligases", "7": "Translocases"}
        
        dfT["EC_class"] = [ec_class[idx[0]] for idx in dfT.index]

        # "class_color" is unused, kept in case a correct way to apply color ranks with update_traces() is found
        dfT["class_color"] = [EC_CLASSES_COLORS[ecclass] for ecclass in dfT["EC_class"]]
        dfT = dfT[["annot_name", "frequency", "EC_class", "class_color"]]
    elif content == 'GO terms':
        dfT = dfT[["annot_name", "frequency"]]

    dfT.sort_values(by=["frequency", "annot_name"], ascending=False, inplace=True)    

    # Count number of EC/GO in each category (left queue, middle, right queue)
    n_9_df = dfT[dfT["frequency"] > 0.9]
    n_1_df = dfT[dfT["frequency"] < 0.1]
    n91_df = dfT[dfT["frequency"].between(0.1,0.9)]
    n_9_an = n_9_df.shape[0]
    n_1_an = n_1_df.shape[0]
    n91_an = n91_df.shape[0]

    # Save lists
    # with open(os.path.join(outpath, "3_analysis", "ecs_last_decile_freqs"), "w") as f:
    #     for idx in n_9_df.index:
    #         f.write(f"{idx}\n")
    # with open(os.path.join(outpath, "3_analysis", "ecs_first_decile_freqs"), "w") as f:
    #     for idx in n_1_df.index:
    #         f.write(f"{idx}\n")
    # with open(os.path.join(outpath, "3_analysis", "ecs_inbetween_deciles_freqs"), "w") as f:
    #     for idx in n91_df.index:
    #         f.write(f"{idx}\n")

    # Fraction of all EC/GO per taxa
    # ------------------------------

    df["fraction"] = df.ne(0).sum(axis=1)/df.shape[1]
    df["taxa"] = df.index
    df.sort_values(by="fraction", ascending=False, inplace=True)
    df = df[["taxa", "fraction"]]

    # Get number of taxa in each category
    n_9_df = df[df["fraction"] > 0.9]
    n_1_df = df[df["fraction"] < 0.1]
    n91_df = df[df["fraction"].between(0.1,0.9)]
    n_9_ob = n_9_df.shape[0]
    n_1_ob = n_1_df.shape[0]
    n91_ob = n91_df.shape[0]

    # Save lists
    # with open(os.path.join(outpath, "3_analysis", "otu_ec_props_last_decile_freqs"), "w") as f:
    #     for idx in n_9_df.index:
    #         f.write(f"{idx}\n")
    # with open(os.path.join(outpath, "3_analysis", "otu_ec_props_first_decile_freqs"), "w") as f:
    #     for idx in n_1_df.index:
    #         f.write(f"{idx}\n")
    # with open(os.path.join(outpath, "3_analysis", "otu_ec_props_inbetween_deciles_freqs"), "w") as f:
    #     for idx in n91_df.index:
    #         f.write(f"{idx}\n")

    data = {
        "df_fractionin_obs": df, 
        "df_annot_frequencies": dfT, 
        "n_9_an": n_9_an, 
        "n_1_an": n_1_an, 
        "n91_an": n91_an, 
        "n_9_ob": n_9_ob, 
        "n_1_ob": n_1_ob, 
        "n91_ob": n91_ob
    }

    return data

def annot_frequencies_in_obs(df, output_folder, content, savefig=True):
    '''
    Plots the frequency of each EC in the taxa (i.e. which EC are in many taxa, and which EC are in few taxa)

        Parameters:
            df (pandas): a pandas df with ECs names, frequencies, and classes as columns (key 'df_ec_frequencies' in the output of create_ec_obs_df())
            output_folder (str) : the path of the output folder
            content (str) : 'GO terms' or 'EC numbers' to specifies which data is processed

        Returns:
            fig : a plotly html figure
    '''
    if content not in ["GO terms", "EC numbers"]:
        raise ValueError("param `content` must either be 'GO terms' or 'EC numbers'")
    
    labels={"annot_name": content, "frequency": "Frequence in modelled taxa"}
    title=f"{content} ordered by their frequency in modelled taxa"

    if content == "EC numbers":
        fig = px.scatter(
            df,
            x = "annot_name",
            y = "frequency",
            color="EC_class",
            height=500,
            category_orders={"annot_name": df.index},
            labels=labels,
            title=title)
    else:
        fig = px.scatter(
            df,
            x = "annot_name",
            y = "frequency",
            height=500,
            category_orders={"annot_name": df.index},
            labels=labels,
            title=title)

    _format_axes(fig)
    fig.add_hline(y=0.9, line_dash="dash", line_width=1, line_color="red")
    fig.add_hline(y=0.1, line_dash="dash", line_width=1, line_color="red")
    fig.add_hrect(y0=0.9, y1=0.1, line_width=0, fillcolor="red", opacity=0.15)

    fig.update_layout(
        yaxis_title=f"Proportion of modelled taxa containing the {content}", 
        plot_bgcolor=PLOT_BGCOLOR, 
        yaxis_range=[-0.1,1.1],
        font=FONTPARAMS)

    if savefig:
        fig.write_html(os.path.join(output_folder, f"{content.replace(' ', '_')}_frequencies_in_taxa.html"))
        write_json(
            fig, 
            os.path.join(output_folder, f"{content.replace(' ', '_')}_frequencies_in_taxa.json"), 
            pretty=True)

    return fig

def fraction_of_all_annot_in_obs(df, df_taxo, output_folder, content, rank="phylum", savefig=True):
    '''
    Plots the fraction of all EC of the dataset in each taxa (i.e. which taxa have many ECs, and which taxa have few ECs)

        Parameters:
            df (pandas) : a pandas df with taxa names, and its fraction of EC as columns (key 'df_ec_fraction' in the output of create_ec_obs_df())
            df_taxo (pandas) : a pandas dataframe containing the taxonomy of taxa (key 'DF_STATS' in the output of swf.post_analysis_config())
            results_path (str) : path to the output folder
            content (str) : 'GO terms' or 'EC numbers' to specifies which data is processed

        Returns:
            fig : a plotly html figure
    '''
    if content not in ["GO terms", "EC numbers"]:
        raise ValueError("param `content` must either be 'GO terms' or 'EC numbers'")

    df = df.join(df_taxo)
    #df.sort_values(by=rank, ascending=False, inplace=True)

    fig = px.scatter(
        df,
        x = "taxa",
        y = "fraction",
        color=rank,
        height=500,
        category_orders={"taxa": df.index},
        labels={"taxa": "Modelled taxa", "fraction": f"Proportion = ({content} in taxa) / (All {content})"},
        title=f"Modelled taxa richness in {content}")

    _format_axes(fig)
    fig.add_hline(y=0.9, line_dash="dash", line_width=1, line_color="red")
    fig.add_hline(y=0.1, line_dash="dash", line_width=1, line_color="red")
    fig.add_hrect(y0=0.9, y1=0.1, line_width=0, fillcolor="red", opacity=0.15)

    fig.update_layout(#yaxis_title=f"Proportion = ({content} in taxa) / (All {content})", 
                      plot_bgcolor=PLOT_BGCOLOR, 
                      yaxis_range=[-0.1,1.1],
                      font=FONTPARAMS)

    if savefig:
        fig.write_html(os.path.join(output_folder, f"{content.replace(' ', '_')}_fraction_per_taxa.html"))
        write_json(
            fig, 
            os.path.join(output_folder, f"{content.replace(' ', '_')}_fraction_per_taxa.json"), 
            pretty=True)

    return fig

def annot_frequencies_in_obs_hist(df, output_folder, content, n_bins=20, savefig=True):
    '''
    For the given rank, plots stacked barplots of how many EC numbers belongs to each taxa of this rank (i.e. which EC are in many taxa, and which EC are in few taxa).

        Parameters:
            df (pandas): a pandas df with EC/GO as rows, taxa as column, filled with the frequencies of EC/GO in taxa
            output_folder (str) : path to the output folder
            content (str) : 'GO terms' or 'EC numbers' to specifies which data is processed

        Returns:
            fig : a plotly html figure
    '''
    if content not in ["GO terms", "EC numbers"]:
        raise ValueError("param `content` must either be 'GO terms' or 'EC numbers'")
    
    labels={"annot_name": content},
    title=f"{content} frequencies in modelled taxa"

    if content == "EC numbers":
        fig = px.histogram(
            df,
            x = "frequency",
            color="EC_class",
            height=500,
            width=500,
            nbins=n_bins,
            labels=labels,
            title=title)
    else:
        fig = px.histogram(
            df,
            x = "frequency",
            height=500,
            width=500,
            nbins=20,
            labels=labels,
            title=title)

    _format_axes(fig)
    _format_trace(fig)
    fig.add_vline(x=0.9, line_dash="dash", line_width=1, line_color="red")
    fig.add_vline(x=0.1, line_dash="dash", line_width=1, line_color="red")
    fig.add_vrect(x0=0.9, x1=0.1, line_width=0, fillcolor="red", opacity=0.15)

    fig.update_layout(
        yaxis_title=f"Number of {content}", 
        xaxis_title=f"Proportion of modelled taxa containing the {content}", 
        plot_bgcolor=PLOT_BGCOLOR, 
        xaxis_range=[-0.1,1.1],
        font=FONTPARAMS)
    
    if savefig:
        fig.write_image(os.path.join(output_folder, f"{content.replace(' ', '_')}_frequencies_in_taxa_hist.pdf"))
        write_json(
            fig, 
            os.path.join(output_folder, f"{content.replace(' ', '_')}_frequencies_in_taxa_hist.json"), 
            pretty=True)

    return fig

def fraction_of_all_annot_in_obs_hist(df, df_taxo, output_folder, content, n_bins=10, rank="phylum", savefig=True):
    '''
    Plots the histogram of the fraction of all EC of the dataset in each taxa (i.e. which taxa have many ECs, and which taxa have few ECs)

        Parameters:
            df (pandas): a pandas df with ECs as rows, taxa as column, filled with the frequencies of ECs in the taxa's clusters
            df_taxo (pandas) : a pandas dataframe containing the taxonomy of taxa (key 'DF_STATS' in the output of swf.post_analysis_config())
            output_folder (str) : path to the output folder
            content (str) : 'GO terms' or 'EC numbers' to specifies which data is processed

        Returns:
            fig : a plotly html figure
    '''
    if content not in ["GO terms", "EC numbers"]:
        raise ValueError("param `content` must either be 'GO terms' or 'EC numbers'")

    df = df.join(df_taxo)
    df.sort_values(by=rank, ascending=False, inplace=True)
    
    fig = px.histogram(
        df,
        x = "fraction",
        color=rank,
        color_discrete_sequence=COLORSCALE,
        height=500,
        width=500,
        nbins=n_bins,
        title=f"Distribution of modelled taxa richness in {content}")

    _format_axes(fig)
    _format_trace(fig)
    fig.add_vline(x=0.9, line_dash="dash", line_width=1, line_color="red")
    fig.add_vline(x=0.1, line_dash="dash", line_width=1, line_color="red")
    fig.add_vrect(x0=0.9, x1=0.1, line_width=0, fillcolor="red", opacity=0.15)

    fig.update_layout(
        xaxis_title=f"Proportion = ({content} in taxa) / (All {content})",
        yaxis_title="Number of modelled taxa", 
        plot_bgcolor=PLOT_BGCOLOR, 
        xaxis_range=[-0.1,1.1],
        font=FONTPARAMS)
    
    if savefig:
        fig.write_image(os.path.join(output_folder, f"{content.replace(' ', '_')}_fraction_per_taxa_hist.pdf"))
        write_json(
            fig, 
            os.path.join(output_folder, f"{content.replace(' ', '_')}_fraction_per_taxa_hist.json"), 
            pretty=True)

    return fig

# ==========================
# EC sunburst summary figure
# ==========================

def ec_sunburst(ec_classes, output_folder, savefig=True):
    '''
    Calls ontosunburst to create a figure summarizing EC numbers categories, counts and proportions
    for the whole dataset

        Parameters:
            ec_classes (list): a list of EC numbers
            output_folder (str): path to the output folder
            savefig (bool): if the figure should be saved (in html format). True by default

        Returns:
            fig (plotly) : a plotly figure object 
    '''
    fig = ontosunburst(interest_set=ec_classes,
        ontology='ec',
        output=os.path.join(output_folder, "ec_classes_sunburst"),
        root_cut="total")
    
    fig.update_layout(
        title="EC numbers categories, counts and proportions", 
        height=1000,
        paper_bgcolor="#ffffff", 
        font_color='#111111', 
        font_size=20)

    fig.update_traces(marker=dict(colorscale=px.colors.diverging.RdYlGn, line_color=LINECOLOR, reversescale=True))

    if savefig:
        fig.write_html(os.path.join(output_folder, "ec_classes_sunburst.html"))
        write_json(
            fig, 
            os.path.join(output_folder, "ec_classes_sunburst.json"), 
            pretty=True)

    return fig


def ec_sunburst_per_model(ec_classes, output_folder, taxgroup, savefig=True):
    '''
    Calls ontosunburst to create a figure summarizing EC numbers categories, counts and proportions
    for a model in the dataset

    Parameters
        ec_classes (list) : a list of EC numbers
        output_folder (str): path to the output folder
        taxgroup (int) : an ete3 taxonomic group id
        savefig (bool): if the figure should be saved (in svg format). True by default

    Returns:
        fig (plotly) : a plotly figure object
    '''

    title = f"{taxgroup}: EC numbers categories, counts and proportions"
    figpath = os.path.join(output_folder, f"{taxgroup}_ec_classes_sunburst")
    jsonpath = os.path.join(output_folder, f"{taxgroup}_ec_classes_sunburst.json")

    fig = ontosunburst(interest_set=ec_classes,
        ontology='ec',
        output=figpath,
        root_cut="total")
    
    fig.update_layout(
        title=title, 
        height=1000,
        width=1000,
        paper_bgcolor="#ffffff", 
        font_color='#111111', 
        font_size=20)

    fig.update_traces(marker=dict(colorscale=px.colors.diverging.RdYlGn, line_color=LINECOLOR, reversescale=True))

    if savefig:
        # fig.write_html(figpath)
        fig.write_image(figpath + '.svg')
        write_json(
            fig, 
            os.path.join(jsonpath), 
            pretty=True)

    return fig

# =========================
# Clustering summary figure
# =========================

def _hex_to_rgb(hex, opacity):
    '''
    Converts an hexadecimal color code to its corresponding rgba value

        Parameters:
            hex (str): an hexadecimal color code ('#' included)
            opacity (float): the opacity value, between 0 (transparent) to 1 (solid)

        Returns:
            rgba (str): a rgba color code : 'rgba(x,y,z,o)'    
    '''
    nohashtag = hex[1:]

    rgba = [int(nohashtag[i:i+2], 16) for i in (0, 2, 4)] + [opacity]
    rgba = [str(i) for i in rgba]
    rgba = f"rgba({','.join(rgba)})"

    return rgba

def data_proteome_representativeness(proteome_tax_id, computed_threshold_folder):
    ''' 
    Prepare data for plotting summary of the clustering step (i.e. representativeness ratio), from the computed threshold folder and the proteomes_tax_id file.

        Parameters:
            pandas_dict (dict): dict of pandas dataframe, with pandas dataframe of proteomes taxa id (key "PROTEOME_TAX_ID" of post_analysis_config() output)
            computed_threshold_folder (str): pathname to computed threshold folder.

        Returns:
            statsdf (pandas) : a pandas dataframe summarizing the mean, max, and min number of protein clusters for each clustering trhesholdand each input taxa
    '''
    # From proteomes_tax_id get the taxonomic rank for each observation name.
    tax_rank = proteome_tax_id['tax_rank'].to_dict()

    tax_obs_name = {key: value for key, value in proteome_tax_id['tax_id_name'].to_dict().items()}

    data = []
    for obs_name in tax_rank:
        taxon_name = tax_obs_name[obs_name]
        tsv_file_path = os.path.join(computed_threshold_folder, taxon_name+'.tsv')
        obs_tsv_file_path = os.path.join(computed_threshold_folder, obs_name+'.tsv')
        # Add a condition for run of esmecata with version under 0.4.0: in computed_threshold folder, files were named according to observation name.
        if not os.path.exists(tsv_file_path) and os.path.exists(obs_tsv_file_path):
            tsv_file_path = obs_tsv_file_path

        full_name = f"Input name : {obs_name}; EsMeCaTa name :{proteome_tax_id.loc[obs_name,'name']}"# More precision for hover info if needed

        tmp_df = pd.read_csv(tsv_file_path, sep='\t')
        for tmp_threshold in np.arange(0, 1.01, 0.025):
            # Compute the number of protein clusters associated with representativeness ratio of tmp_threshold.
            nb_protein_cluster_ratio = len(tmp_df[tmp_df['cluster_ratio']>= tmp_threshold])
            data.append([obs_name, full_name, tax_rank[obs_name], tmp_threshold, nb_protein_cluster_ratio])

    df = pd.DataFrame(data, columns=['obs_name', 'full_name', 'rank', 'clust', 'count'])

    return df

def create_proteome_representativeness_lineplot_px(df, clust_threshold, output_folder, savefig=True):
    '''     
    This figure shows the representativeness ratio and the number of associated protein clusters according to the taxonomic rank.

        Parameters:
            df (pandas) : a dataframe summarizing proteomes by cluster and by threshold, output of data_proteome_representativeness()
            clust_threshold (float) : the threshold between 0 and 1 fixed by the user for the clustering process of EsMeCaTa
            output_folder (str): pathname to the output figure file.

        Returns:
            fig : a plotly html figure
    '''
    #statsdf = df.groupby(['rank'])['count'].agg(['mean', 'count', 'std'])
    #statsdfs_detail = df.groupby(['obs_name'])

    # Compute confidence interval
    # h = []
    # l = []
    # for rank, row in statsdf.iterrows():
    #     ltmp, htmp = st.t.interval(0.95, row["count"]-1, loc=row["mean"], scale=row["std"])
    #     h.append(htmp)
    #     l.append(ltmp)    
    # statsdf["ciho"] = h
    # statsdf["cilo"] = l

    fig = go.Figure()

    for rank in df["rank"].unique(): # df["rank"].unique() or statsdf.index
        rank_df = df.loc[df["rank"] == rank]
        rank_df = rank_df.groupby(["clust"])["count"].agg(["mean", "min", "max"])

        # tmp_df["lo"] = tmp_df["count"] - statsdf.loc[rank, "cilo"]
        # tmp_df["hi"] = tmp_df["count"] + statsdf.loc[rank, "ciho"]
        fig.add_traces([
            go.Scatter(
                name=f"{rank} mean" ,
                x=rank_df.index,
                y=rank_df["mean"],
                mode="lines+markers",
                line=dict(color=RANK2COL[rank]),
                hovertemplate = "Threshold: %{x}; Clusters: %{y}"),
            go.Scatter(
                name=f"{rank} min",
                x=rank_df.index,
                y=rank_df["min"],
                mode="lines",
                line=dict(width=0, color=RANK2COL[rank]),
                hoverinfo='skip',
                showlegend=False),
            go.Scatter(
                name=f"{rank} max",
                x=rank_df.index,
                y=rank_df["max"],
                mode="lines",
                fill='tonexty',
                fillcolor=_hex_to_rgb(RANK2COL[rank], 0.2),
                line=dict(width=0, color=RANK2COL[rank]),
                hoverinfo='skip',
                showlegend=False)
        ])

    fig.add_vline(
        x=clust_threshold, 
        line_dash="dash", 
        line_width=1,
        line_color="red", 
        annotation_text="EsMeCaTa's clustering threshold")

    _format_axes(fig)
    fig.update_xaxes(range=[0, 1])  
    fig.update_yaxes(type="log")    

    fig.update_layout(
        title="Count of retained clusters of proteins (split by by taxonomic rank) according to the clustering threshold",
        xaxis_title="Clustering threshold (0 : pan-proteome, 1 : core-proteome)", 
        yaxis_title="Number of clusters of proteomes retained", 
        height=800, 
        plot_bgcolor=PLOT_BGCOLOR,
        legend=dict(x=1, y=1, xanchor='right', yanchor='top'),
        font=FONTPARAMS)

    if savefig:
        fig.write_image(os.path.join(output_folder, 'proteome_representativeness.pdf'))
        write_json(
            fig, 
            os.path.join(output_folder, 'proteome_representativeness.json'), 
            pretty=True)

    return fig

def proteomes_representativeness_details(df, clust_threshold, output_folder, savefig=True):
    '''
    This figure shows details of the representativeness ratio and the number of associated protein clusters according to each cluster. There is one sub-pnale by taxonomic rank.

        Parameters:
            df (pandas) : a dataframe summarizing proteomes by cluster and by threshold, output of data_proteome_representativeness().
            clust_threshold (float) : the threshold between 0 and 1 fixed by the user for the clustering process of EsMeCaTa.
            output_folder (str): pathname to the output figure file.
            savefig (bool): boolean to create output files.

        Returns:
            fig : a plotly html figure
    '''

    ncbi_ranks = ["clade", "superkingdom","kingdom","subkingdom","superphylum","phylum",
        "subphylum","infraphylum","superclass","class","subclass",
        "infraclass","cohort","subcohort","superorder","order",
        "suborder","infraorder","parvorder","superfamily","family",
        "subfamily","tribe","subtribe","genus","subgenus",
        "section","subsection","series","subseries","species group",
        "species subgroup","species","forma specialis","subspecies","varietas",
        "subvariety","forma","serogroup","serotype","strain","isolate"]
    
    ranks = df["rank"].unique()
    ncbi_ranks = [rank for rank in ncbi_ranks if rank in ranks]

    titles=[f"Details for {rank}" for rank in ncbi_ranks]

    n = df["rank"].nunique()
    fig = make_subplots(
        rows=n, 
        cols=1,
        shared_xaxes=False,
        vertical_spacing=0.075,
        subplot_titles=titles)  

    i = 1
    # for rank in df["rank"].unique():
    for rank in ncbi_ranks:

        rank_df = df.loc[df["rank"] == rank]        

        for input_taxa in rank_df["obs_name"].unique():
            rank_df_taxa = rank_df.loc[df["obs_name"] == input_taxa]

            fname = rank_df_taxa["full_name"].iloc[0]

            fig.add_trace(
                go.Scatter(
                    name=fname, # input_taxa
                    x=rank_df_taxa["clust"],
                    y=rank_df_taxa["count"],
                    mode="lines+markers",
                    line=dict(color=RANK2COL[rank]),
                    showlegend=False,
                    hovertemplate = "Threshold: %{x}; Clusters: %{y}"
                ),
                row=i,
                col=1
            )
        i += 1

    fig.add_vline(x=clust_threshold, 
        line_dash="dash", 
        line_width=1, 
        line_color="red", 
        annotation_text="EsMeCaTa's clustering threshold")
        
    _format_axes(fig)    
    fig.update_yaxes(type="log")
    fig.update_xaxes(range=[0, 1])
        
    fig.update_layout(#title=f"Details for {rank} clusters",
        xaxis_title="Clustering threshold (0 : pan-proteome, 1 : core-proteome)", 
        yaxis_title="Number of clusters of proteomes retained", 
        height=n*400,
        width=500,
        plot_bgcolor=PLOT_BGCOLOR,
        font=FONTPARAMS)
        
    i = 1
    for rank in df["rank"].unique():
        fig.update_xaxes(title_text="Clustering threshold (0 : pan-proteome, 1 : core-proteome)", row=i, col=1)
        fig.update_yaxes(title_text="Number of clusters of proteomes retained", row=i, col=1)
        i += 1
        
    if savefig:
        fig.write_html(os.path.join(output_folder, 'proteome_representativeness_details.html'))
        write_json(
            fig, 
            os.path.join(output_folder, 'proteome_representativeness_details.json'),
            pretty=True)

    return fig

# ======================
# Metadata Page function
# ======================

def reproducibility_tokens(outdir):
    '''
    Loads json metadata of all the EsmeCata workflow and converts it to a string.

        Parameters:
            outdir (str) : path to the results of an EsMeCaTa workflow

        Returns:
            metadata (str) : all metadata of the workflow
    '''
    reprometa_proteomes = json.load(open(os.path.join(outdir, '0_proteomes', 'esmecata_metadata_proteomes.json'), 'r'))
    reprometa_clustering = json.load(open(os.path.join(outdir, '1_clustering', 'esmecata_metadata_clustering.json'), 'r'))
    reprometa_annotation = json.load(open(os.path.join(outdir, '2_annotation', 'esmecata_metadata_annotation.json'), 'r'))

    metadata = {"proteomes": reprometa_proteomes,
         "clustering": reprometa_clustering,
         "annotation": reprometa_annotation}
    
    # metadata = json.dumps(metadata, indent=4)

    return metadata



# Copyright (C) 2023-2025 Alice Mataigne - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>
""" 
Uses esmecata sunburst on lists of EC, one per taxonomic rank.
The output of the function `create_dataset_annotation_file()` is 
needed (in its default path)
"""

import os
import pandas as pd
from esmecata.report.stats_workflow_figures import ec_sunburst_per_model


def create_sunburst_ec(esmecata_output_folder, output_folder):
    dataset_annotation_ec_file = os.path.join(output_folder, 'dataset_annotation_ec.tsv')
    proteome_tax_id_filepath = os.path.join(esmecata_output_folder, '0_proteomes', 'proteome_tax_id.tsv')
    # Get EC data of biogas dataset and taxonomic lineages
    df_ec = pd.read_table(dataset_annotation_ec_file,
        header=0,
        index_col=0,
        sep='\t')

    # Used to group inputs with the same esmecata output (same rank = same annotation)
    taxo = pd.read_table(proteome_tax_id_filepath,
        header=0,
        index_col=0,
        sep='\t')

    if not os.path.exists(os.path.join(output_folder, 'ec_sunburst_per_model')):
        os.mkdir(os.path.join(output_folder, 'ec_sunburst_per_model'))

    # observations_and_groups = {}

    for taxid in taxo["name"].unique():
        group = taxo[taxo["name"] == taxid].index
        # observations_and_groups[taxid] = group
        df_ec_reduced = df_ec.loc[group,]
        ecs = df_ec_reduced.loc[:, (df_ec_reduced != 0).all()].columns.tolist()
        fig = ec_sunburst_per_model(ecs, output_folder, taxid, True)



# Copyright (C) 2021-2025 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import argparse
import logging
import os
import sys
import time

from esmecata.report.workflow_create_report import run_create_workflow_report, run_proteomes_report_creation, run_clustering_report_creation, \
                                                                    run_annotation_report_creation
from esmecata.utils import is_valid_dir
from esmecata import __version__ as VERSION

MESSAGE = '''
Create report files from esmecata output folder.
'''
REQUIRES = '''
Requires: datapane, plotly, kaleido, ontosunburst.
'''

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

def main():
    start_time = time.time()

    parser = argparse.ArgumentParser(
        'esmecata_report',
        description=MESSAGE + ' For specific help on each subcommand use: esmecata {cmd} --help',
        epilog=REQUIRES
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + VERSION + '\n')

    parent_parser_i = argparse.ArgumentParser(add_help=False)
    parent_parser_i.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='EsMeCaTa input file.',
        metavar='INPUT_FILE')

    parent_parser_f = argparse.ArgumentParser(add_help=False)
    parent_parser_f.add_argument(
        '-f',
        '--folder',
        dest='input_folder',
        required=True,
        help='EsMeCaTa output folder.',
        metavar='INPUT_FOLDER')

    parent_parser_o = argparse.ArgumentParser(add_help=False)
    parent_parser_o.add_argument(
        '-o',
        '--output',
        dest='output',
        required=True,
        help='Output directory path.',
        metavar='OUPUT_DIR')

    parent_parser_svg = argparse.ArgumentParser(add_help=False)
    parent_parser_svg.add_argument(
        '--svg',
        dest='create_svg',
        action='store_true',
        required=False,
        help='Create SVG files of picture.')

    # subparsers
    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands:',
        dest='cmd')

    create_report_parser = subparsers.add_parser(
        'create_report',
        help='Create report from esmecata output folder of workflow or precomputed subcommands.',
        parents=[
            parent_parser_i, parent_parser_f, parent_parser_o, parent_parser_svg
            ],
        allow_abbrev=False)

    create_report_proteomes_parser = subparsers.add_parser(
        'create_report_proteomes',
        help='Create report from esmecata output folder of proteomes subcommand.',
        parents=[
            parent_parser_f, parent_parser_o, parent_parser_svg
            ],
        allow_abbrev=False)

    create_report_clustering_parser = subparsers.add_parser(
        'create_report_clustering',
        help='Create report from esmecata output folder of clustering subcommand.',
        parents=[
            parent_parser_f, parent_parser_o, parent_parser_svg
            ],
        allow_abbrev=False)

    create_report_annotation_parser = subparsers.add_parser(
        'create_report_annotation',
        help='Create report from esmecata output folder of annotation subcommand.',
        parents=[
            parent_parser_f, parent_parser_o, parent_parser_svg
            ],
        allow_abbrev=False)

    args = parser.parse_args()

    # If no argument print the help.
    if len(sys.argv) == 1 or len(sys.argv) == 0:
        parser.print_help()
        sys.exit(1)

    is_valid_dir(args.output)

    # add logger in file
    formatter = logging.Formatter('%(message)s')
    log_file_path = os.path.join(args.output, f'esmecata_analysis_{args.cmd}.log')
    file_handler = logging.FileHandler(log_file_path, 'w+')
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # set up the default console logger
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if args.cmd == 'create_report':
        run_create_workflow_report(args.input, args.input_folder, args.output, args.create_svg)
    elif args.cmd == 'create_report_proteomes':
        run_proteomes_report_creation(args.input_folder, args.output, args.create_svg)
    elif args.cmd == 'create_report_clustering':
        run_clustering_report_creation(args.input_folder, args.output, args.create_svg)
    elif args.cmd == 'create_report_annotation':
        run_annotation_report_creation(args.input_folder, args.output, args.create_svg)

    logger.info("--- Total runtime %.2f seconds ---" % (time.time() - start_time))
    logger.warning(f'--- Logs written in {log_file_path} ---')


if __name__ == '__main__':
    main()



# Copyright (C) 2021-2025 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import gzip
import json
import logging
import os
import shutil
import subprocess
import sys
import time

from Bio import SeqIO
from Bio import __version__ as biopython_version
from shutil import which

from esmecata import __version__ as esmecata_version
from esmecata.utils import is_valid_path, is_valid_dir

logger = logging.getLogger(__name__)


def compute_stat_clustering(output_folder, stat_file=None):
    """Compute stat associated to the number of proteome for each taxonomic affiliations.

    Args:
        output_folder (str): pathname to the result folder containing mmseqs results
        stat_file (str): pathname to the tsv stat file

    Returns:
        clustering_numbers (dict): dict containing observation names (as key) associated with the number of protein clusters
    """
    clustering_taxon_id_file = os.path.join(output_folder, 'proteome_tax_id.tsv')
    result_folder = os.path.join(output_folder, 'computed_threshold')

    proteomes_taxa_id_names = get_proteomes_tax_id_name(clustering_taxon_id_file)

    tax_name_clustering_numbers = {}
    for clustering_file in os.listdir(result_folder):
        clustering_file_path = os.path.join(result_folder, clustering_file)
        with open(clustering_file_path, 'r') as open_clustering_file_path:
            csvreader = csv.reader(open_clustering_file_path, delimiter='\t')
            next(csvreader)
            cluster_0 = []
            cluster_0_5 = []
            cluster_0_95 = []
            for line in csvreader:
                protein_cluster_threshold = float(line[1])
                if protein_cluster_threshold >= 0.95:
                    cluster_0_95.append(line[0])
                if protein_cluster_threshold >= 0.5:
                    cluster_0_5.append(line[0])
                if protein_cluster_threshold >= 0:
                    cluster_0.append(line[0])

        tax_name_clustering_numbers[clustering_file.replace('.tsv', '')] = [cluster_0, cluster_0_5, cluster_0_95]

    clustering_numbers = {}
    for observation_name in proteomes_taxa_id_names:
        tax_name = proteomes_taxa_id_names[observation_name]
        clustering_numbers[observation_name] = tax_name_clustering_numbers[tax_name]

    if stat_file:
        with open(stat_file, 'w') as stat_file_open:
            csvwriter = csv.writer(stat_file_open, delimiter='\t')
            csvwriter.writerow(['observation_name', 'Number_protein_clusters_panproteome', 'Number_protein_clusters_kept', 'Number_protein_clusters_coreproteome'])
            for observation_name in clustering_numbers:
                cluster_0 = len(clustering_numbers[observation_name][0])
                cluster_0_5 = len(clustering_numbers[observation_name][1])
                cluster_0_95 = len(clustering_numbers[observation_name][2])
                csvwriter.writerow([observation_name, cluster_0, cluster_0_5, cluster_0_95])

    return clustering_numbers


def get_proteomes_tax_id_name(proteomes_tax_id_file_path, header_to_extract='tax_id_name'):
    """ Extract tax_name + tax_id associated with observation name.

    Args:
        proteomes_tax_id_file_path (str): pathname to the proteomes_tax_id file.
        header_to_extract (str): name of the header to extract data from (by default tax_id_name).

    Returns:
        proteomes_taxa_id_names (dict): dict containing observation names (as key) associated with tax name + tax_id used for proteomes (as value)
    """
    proteomes_taxa_id_names = {}
    with open(proteomes_tax_id_file_path, 'r') as proteome_tax_file:
        csvreader = csv.DictReader(proteome_tax_file, delimiter='\t')
        for line in csvreader:
            observation_name = line['observation_name']
            data_extract = line[header_to_extract]
            proteomes_taxa_id_names[observation_name] = data_extract

    return proteomes_taxa_id_names


def copy_already_clustered_file(output_folder, already_clustered_obs_name, new_observation_name, file_extension='.tsv'):
    """ Copy files from an observation name with the same tax ID than the new ones.
    Args:
        output_folder (str): path to the ouput result folder
        already_clustered_obs_name (str): name of the already clustered observation name
        new_observation_name (str): name of the new observation name
        file_extension (str): file extension (by default '.tsv')
    """
    already_cluster_output_file = os.path.join(output_folder, already_clustered_obs_name + file_extension)
    cluster_output_file = os.path.join(output_folder, new_observation_name + file_extension)
    shutil.copyfile(already_cluster_output_file, cluster_output_file)


def run_mmseqs(observation_name, observation_name_proteomes, mmseqs_tmp_path, nb_core, mmseqs_options, linclust):
    """Run MMseqs2 on proteomes for an observation name

    Args:
        observation_name (str): observation name associated to a taxonomic affiliations
        observation_name_proteomes (list): list of pathname to each proteomes associated to the observation_name
        mmseqs_tmp_path (str): pathname to the folder which will contain mmseqs results
        nb_core (int): number of CPU-cores for mmseqs
        mmseqs_options (str): options that will be used by mmseqs (default: '--min-seq-id', '0.3', '-c', '0.8')
        linclust (bool): use linclust for faster clustering

    Returns:
        mmseqs_tmp_clustered_tabulated (str): pathname to the mmseqs tabulated file containg protein cluster
        mmseqs_tmp_representative_fasta (str): pathname to the mmseqs fasta file containing representative protein sequences
        mmseqs_consensus_fasta (str): pathname to the mmseqs fasta file containing consensus protein sequences
    """
    mmseqs_tmp_cluster = os.path.join(mmseqs_tmp_path, observation_name)
    is_valid_dir(mmseqs_tmp_cluster)

    mmseqs_tmp_cluster_output = os.path.join(mmseqs_tmp_cluster, 'cluster')
    mmseqs_tmp_clustered_tabulated = mmseqs_tmp_cluster_output+'_cluster.tsv'
    mmseqs_tmp_representative_fasta = mmseqs_tmp_cluster_output + '_rep_seq.fasta'
    mmseqs_consensus_fasta = mmseqs_tmp_cluster_output + '_con_seq.fasta'
    # Run mmmseqs to find protein clusters.

    # Code using mmseqs database and mmseqs modules to cluster instead of easy-cluster.
    mmseqs_tmp_db = os.path.join(mmseqs_tmp_cluster, 'db')
    mmseqs_tmp_db_clustered = os.path.join(mmseqs_tmp_cluster, 'cluster_db')
    mmseqs_tmp_cluster_tmp = os.path.join(mmseqs_tmp_cluster, 'cluster_tmp')
    mmseqs_seq_db =  os.path.join(mmseqs_tmp_cluster, 'cluster_seq')
    mmseqs_profile =  os.path.join(mmseqs_tmp_cluster, 'cluster_profile')
    mmseqs_consensus =  os.path.join(mmseqs_tmp_cluster, 'cluster_consensus')

    if not os.path.exists(mmseqs_tmp_clustered_tabulated):
        # Create database containing the protein sequences from all the proteomes of a taxon.
        subprocess.call(['mmseqs', 'createdb', *observation_name_proteomes, mmseqs_tmp_db, '-v', '2'])

        # Cluster the protein sequences.

        cluster_cmd = ['mmseqs']
        if linclust:
            cluster_cmd += ['linclust']
        else:
            cluster_cmd += ['cluster']

        cluster_cmd += [mmseqs_tmp_db, mmseqs_tmp_db_clustered, mmseqs_tmp_cluster_tmp, '--threads', str(nb_core), '-v', '2']

        # If no option given by the user, use the default options: '--min-seq-id', '0.3', '-c', '0.8'.
        # Sequence identity of 30% and coverage of 80%.
        if not mmseqs_options:
            cluster_cmd += ['--min-seq-id', '0.3', '-c', '0.8']
        else:
            cluster_cmd += mmseqs_options.split(' ')

        subprocess.call(cluster_cmd)

        # Create sequence database with representative from the clustered proteins.
        subprocess.call(['mmseqs', 'createsubdb', mmseqs_tmp_db_clustered, mmseqs_tmp_db, mmseqs_seq_db, '-v', '2'])
        # Create the profile from the clustering.
        subprocess.call(['mmseqs', 'result2profile', mmseqs_seq_db, mmseqs_tmp_db, mmseqs_tmp_db_clustered, mmseqs_profile, '--threads', str(nb_core), '-v', '2'])
        # Create the consensus from the profile.
        subprocess.call(['mmseqs', 'profile2consensus', mmseqs_profile, mmseqs_consensus, '--threads', str(nb_core), '-v', '2'])
        # Create the consensus fasta file.
        subprocess.call(['mmseqs', 'convert2fasta', mmseqs_consensus, mmseqs_consensus_fasta, '-v', '2'])
        # Create TSV resulting files to be analysed after.
        subprocess.call(['mmseqs', 'createtsv', mmseqs_tmp_db, mmseqs_tmp_db, mmseqs_tmp_db_clustered, mmseqs_tmp_clustered_tabulated, '--threads', str(nb_core), '-v', '2'])
        # Create fasta file containing representative proteins (representatives are the first protein of the alignment).
        subprocess.call(['mmseqs', 'convert2fasta', mmseqs_seq_db, mmseqs_tmp_representative_fasta, '-v', '2'])

    # Old clustering method used with easy-cluster.
    """
    if not os.path.exists(mmseqs_tmp_clustered_tabulated):
        # Code using easy-cluster to extract representative protein.
        subprocess.call(['mmseqs', 'easy-cluster', *cluster_fasta_files[cluster], mmseqs_tmp_cluster_output, mmseqs_tmp_cluster, '--threads', str(nb_core), '-v', '2', '--min-seq-id', '0.3'])
    """

    return mmseqs_tmp_clustered_tabulated, mmseqs_tmp_representative_fasta, mmseqs_consensus_fasta


def extrat_protein_cluster_from_mmseqs(mmseqs_tmp_clustered_tabulated, cluster_proteomes_output_file):
    """Extract protein cluster from mmseqs into a dictionary

    Args:
        mmseqs_tmp_clustered_tabulated (str): pathname to the mmseqs tabulated file containg protein cluster
        cluster_proteomes_output_file (str): pathname to the output file showing the content of each protein cluster

    Returns:
        protein_clusters (dict): protein clusters found by mmseqs (representative protein as key and all the protein in the cluster as value)
    """
    # Extract protein clusters in a dictionary.
    # The representative protein is the key and all the proteins in the cluster are the values.
    protein_clusters = {}
    with open(mmseqs_tmp_clustered_tabulated, 'r') as input_file:
        csvreader = csv.reader(input_file, delimiter='\t')
        for row in csvreader:
            if row[0] not in protein_clusters:
                protein_clusters[row[0]] = [row[1]]
            else:
                protein_clusters[row[0]].append(row[1])

    with open(cluster_proteomes_output_file, 'w') as output_file:
        csvwriter = csv.writer(output_file, delimiter='\t')
        for rep_protein in protein_clusters:
            csvwriter.writerow([rep_protein, *[prot for prot in protein_clusters[rep_protein]]])

    return protein_clusters


def compute_proteome_representativeness_ratio(protein_clusters, observation_name_proteomes, computed_threshold_file=None):
    """Compute for each protein cluster the ratio of representation of each proteomes in the cluster.

    Args:
        protein_clusters (dict): protein clusters found by mmseqs (representative protein as key and all the protein in the cluster as value)
        observation_name_proteomes (list): list of pathname to each proteomes associated to the observation_name
        computed_threshold_file (str): pathname to the output file containing the computed ratio

    Returns:
        number_proteomes (int): number of the proteomes for the observation_name
        rep_prot_organims (dict): names of each proteomes associated with each protein cluster
        computed_threshold_cluster (dict): ratio of proteomes representativeness for each protein cluster
    """
    # Retrieve protein ID and the corresponding proteome.
    organism_prots = {}
    for fasta_file in observation_name_proteomes:
        with gzip.open(fasta_file, 'rt') as fasta_handle:
            for record in SeqIO.parse(fasta_handle, 'fasta'):
                compressed_filebasename = os.path.basename(fasta_file)
                fasta_filebasename = os.path.splitext(compressed_filebasename)[0]
                filebasename = os.path.splitext(fasta_filebasename)[0]
                if '|' in record.id:
                    prot_id = record.id.split('|')[1]
                else:
                    prot_id = record.id
                organism_prots[prot_id] = filebasename

    number_proteomes = len(observation_name_proteomes)

    # Compute the ratio between the number of proteome represented by a protein in the cluster and the total number of proteome.
    rep_prot_organims = {}
    computed_threshold_cluster = {}
    for rep_protein in protein_clusters:
        rep_prot_organims[rep_protein] = set([organism_prots[prot] for prot in protein_clusters[rep_protein]])
        computed_threshold_cluster[rep_protein] = len(rep_prot_organims[rep_protein]) / number_proteomes

    if computed_threshold_file:
        # Create a tsv file containing the computer threshold (number of organism in the cluster compared to the total number of organism) for each organisms.
        with open(computed_threshold_file, 'w') as output_file:
            csvwriter = csv.writer(output_file, delimiter='\t')
            csvwriter.writerow(['representative_protein', 'cluster_ratio', 'proteomes'])
            for rep_protein in rep_prot_organims:
                csvwriter.writerow([rep_protein, computed_threshold_cluster[rep_protein], ','.join(rep_prot_organims[rep_protein])])

    return number_proteomes, rep_prot_organims, computed_threshold_cluster


def filter_protein_cluster(protein_clusters, number_proteomes, rep_prot_organims, computed_threshold_cluster,
                           clust_threshold, cluster_proteomes_filtered_output_file=None):
    """Filter protein cluster according to the representation of each proteomes in the cluster.

    Args:
        protein_clusters (dict): protein clusters found by mmseqs (representative protein as key and all the protein in the cluster as value)
        number_proteomes (int): number of the proteomes for the observation_name
        rep_prot_organims (dict): names of each proteomes associated with each protein cluster
        computed_threshold_cluster (dict): ratio of proteomes representativeness for each protein cluster
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        cluster_proteomes_filtered_output_file (str): pathname to the output file showing the protein cluster kept after filtering

    Returns:
        protein_cluster_to_keeps (list): list containing representative protein IDs associated with cluster that are kept
    """
    # Keep a protein cluster according to the ratio and create a list of representative proteins (being the protein cluster to keep).
    protein_cluster_to_keeps = []
    reference_threshold = (clust_threshold * number_proteomes) / number_proteomes

    for rep_protein in rep_prot_organims:
        if computed_threshold_cluster[rep_protein] >= reference_threshold:
            protein_cluster_to_keeps.append(rep_protein)

    if cluster_proteomes_filtered_output_file:
        with open(cluster_proteomes_filtered_output_file, 'w') as filtered_output_file:
            filtered_csvwriter = csv.writer(filtered_output_file, delimiter='\t')
            for rep_protein in protein_cluster_to_keeps:
                filtered_csvwriter.writerow([rep_protein, *[prot for prot in protein_clusters[rep_protein]]])

    # Use set for faster search using 'in'.
    protein_cluster_to_keeps = set(protein_cluster_to_keeps)

    return protein_cluster_to_keeps


def make_clustering(proteome_folder, output_folder, nb_core, clust_threshold, mmseqs_options, linclust, remove_tmp):
    """From the proteomes found by esmecata proteomes, create protein cluster for each taxonomic affiliations.

    Args:
        proteome_folder (str): pathname to folder from esmecata folder
        output_folder (str): pathname to the output folder
        nb_core (int): number of CPU-cores to be used by mmseqs
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        mmseqs_options (str): use alternative mmseqs option
        linclust (bool): use linclust
        remove_tmp (bool): remove the tmp files
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|clustering| Begin clustering.')

    # Check if mmseqs is in path.
    mmseqs_path = which('mmseqs')
    if not mmseqs_path:
        logger.critical('|EsMeCaTa|clustering| mmseqs not available in path, esmecata will not be able to cluster the proteomes.')
        sys.exit(1)

    if not is_valid_dir(proteome_folder):
        logger.critical('|EsMeCaTa|clustering| Input must be a folder %s.', proteome_folder)
        sys.exit(1)

    # Use the proteomes folder created by retrieve_proteome.py.
    proteome_tax_id_pathname = os.path.join(proteome_folder, 'proteome_tax_id.tsv')

    if not is_valid_path(proteome_tax_id_pathname):
        logger.critical(f"|EsMeCaTa|clustering| Missing output from esmecata proteomes in {proteome_tax_id_pathname}.")
        sys.exit(1)

    reference_proteins_path = os.path.join(output_folder, 'reference_proteins')
    is_valid_dir(reference_proteins_path)

    cluster_founds_path = os.path.join(output_folder, 'cluster_founds')
    is_valid_dir(cluster_founds_path)

    computed_threshold_path = os.path.join(output_folder, 'computed_threshold')
    is_valid_dir(computed_threshold_path)

    already_performed_clustering = [reference_protein_file.replace('.tsv', '') for reference_protein_file in os.listdir(reference_proteins_path)]

    # Create a dictionary with observation_name as key and the pathname to the proteomes associated to this observation_name as value.
    observation_name_fasta_files = {}
    with open(proteome_tax_id_pathname, 'r') as proteome_tax_file:
        csvreader = csv.DictReader(proteome_tax_file, delimiter='\t')
        for line in csvreader:
            proteomes = line['proteome'].split(',')
            tax_id_name = line['tax_id_name']
            proteomes_path = [os.path.join(proteome_folder, 'proteomes', proteome+'.faa.gz') for proteome in proteomes]
            if tax_id_name not in already_performed_clustering:
                observation_name_fasta_files[tax_id_name] = proteomes_path
            else:
                logger.info('|EsMeCaTa|clustering| Already performed clustering for %s.', tax_id_name)

    is_valid_dir(output_folder)

    # Create metadata file.
    clustering_metadata = {}
    clustering_metadata['tool_options'] = {'proteome_folder': proteome_folder, 'output_folder': output_folder, 'nb_core':nb_core,
                                        'clust_threshold':clust_threshold, 'mmseqs_options': mmseqs_options, 'linclust':linclust,
                                        'remove_tmp': remove_tmp}

    clustering_metadata['tool_dependencies'] = {}
    subprocess_output = subprocess.check_output(['mmseqs', 'version'])
    mmseqs_version = subprocess_output.decode('utf-8')
    clustering_metadata['tool_dependencies']['mmseqs_version'] = mmseqs_version
    clustering_metadata['tool_dependencies']['mmseqs_path'] = mmseqs_path
    clustering_metadata['tool_dependencies']['python_package'] = {}
    clustering_metadata['tool_dependencies']['python_package']['Python_version'] = sys.version
    clustering_metadata['tool_dependencies']['python_package']['biopython'] = biopython_version
    clustering_metadata['tool_dependencies']['python_package']['esmecata'] = esmecata_version

    # Create tmp folder for mmseqs analysis.
    mmseqs_tmp_path = os.path.join(output_folder, 'mmseqs_tmp')
    is_valid_dir(mmseqs_tmp_path)

    # Create output folder containing shared representative proteins.
    reference_proteins_representative_fasta_path = os.path.join(output_folder, 'reference_proteins_representative_fasta')
    is_valid_dir(reference_proteins_representative_fasta_path)

    reference_proteins_consensus_fasta_path = os.path.join(output_folder, 'reference_proteins_consensus_fasta')
    is_valid_dir(reference_proteins_consensus_fasta_path)

    proteome_taxon_id_file = os.path.join(proteome_folder, 'proteome_tax_id.tsv')
    clustering_taxon_id_file = os.path.join(output_folder, 'proteome_tax_id.tsv')

    if os.path.exists(clustering_taxon_id_file):
        if not os.path.samefile(proteome_taxon_id_file, clustering_taxon_id_file):
            os.remove(clustering_taxon_id_file)
            shutil.copyfile(proteome_taxon_id_file, clustering_taxon_id_file)
    else:
        shutil.copyfile(proteome_taxon_id_file, clustering_taxon_id_file)

    proteomes_taxa_names = get_proteomes_tax_id_name(proteome_taxon_id_file)

    all_tax_names = set(list(proteomes_taxa_names.values())) - set(already_performed_clustering)
    nb_taxa_names = len(all_tax_names)
    logger.info('|EsMeCaTa|clustering| Create consensus proteomes for %d taxa.', nb_taxa_names)

    # For each OTU run mmseqs easy-cluster on them to found the clusters that have a protein in each proteome of the OTU.
    # We take the representative protein of a cluster if the cluster contains a protein from all the proteomes of the OTU.
    # If this condition is not satisfied the cluster will be ignored.
    # Then a fasta file containing all the representative proteins for each OTU is written in representative_fasta folder.
    for index, proteomes_tax_name in enumerate(all_tax_names):
        # Get proteomes associated with taxon name.
        observation_name_proteomes = observation_name_fasta_files[proteomes_tax_name]

        # Change space with '_' to avoid issue.
        proteomes_tax_name = proteomes_tax_name
        # If the computed threshold file exists, mmseqs has already been run.
        mmseqs_tmp_cluster = os.path.join(mmseqs_tmp_path, proteomes_tax_name)
        # Run mmseqs on organism.
        # Delete previous mmseqs2 run if it exists to avoid overwritting issues.
        if os.path.exists(mmseqs_tmp_cluster):
            shutil.rmtree(mmseqs_tmp_cluster)
        mmseqs_tmp_clustered_tabulated, mmseqs_tmp_representative_fasta, mmseqs_consensus_fasta = run_mmseqs(proteomes_tax_name, observation_name_proteomes, mmseqs_tmp_path, nb_core,
                                                                                                             mmseqs_options, linclust)

        # Extract protein clusters from mmseqs results.
        cluster_proteomes_output_file = os.path.join(cluster_founds_path, proteomes_tax_name+'.tsv')
        protein_clusters = extrat_protein_cluster_from_mmseqs(mmseqs_tmp_clustered_tabulated, cluster_proteomes_output_file)

        # Compute proteome representativeness ratio.
        computed_threshold_file = os.path.join(computed_threshold_path, proteomes_tax_name+'.tsv')
        number_proteomes, rep_prot_organims, computed_threshold_cluster = compute_proteome_representativeness_ratio(protein_clusters,
                                                                                                                    observation_name_proteomes, computed_threshold_file)

        # Filter protein cluster for each protein cluster.
        cluster_proteomes_filtered_output_file = os.path.join(reference_proteins_path, proteomes_tax_name+'.tsv')
        protein_cluster_to_keeps = filter_protein_cluster(protein_clusters, number_proteomes, rep_prot_organims, computed_threshold_cluster,
                                                        clust_threshold, cluster_proteomes_filtered_output_file)

        logger.info('|EsMeCaTa|clustering| %d protein clusters kept for %s (%d on %d).', len(protein_cluster_to_keeps), proteomes_tax_name, index+1, nb_taxa_names)

        # Create BioPython records with the representative proteins kept.
        new_records = [record for record in SeqIO.parse(mmseqs_tmp_representative_fasta, 'fasta')
                                        if ('|' in record.id and record.id.split('|')[1] in protein_cluster_to_keeps) or (record.id in protein_cluster_to_keeps)]

        # Do not create fasta file when 0 sequences were kept.
        if len(new_records) > 0:
            # Create output proteome file for OTU.
            representative_fasta_file = os.path.join(reference_proteins_representative_fasta_path, proteomes_tax_name+'.faa')
            SeqIO.write(new_records, representative_fasta_file, 'fasta')
        else:
            logger.info('|EsMeCaTa|clustering| 0 protein clusters %s, no representatitve fasta created (%d on %d).', proteomes_tax_name, index+1, nb_taxa_names)
        del new_records

        # Create BioPython records with the consensus proteins kept.
        consensus_new_records = [record for record in SeqIO.parse(mmseqs_consensus_fasta, 'fasta')
                                        if ('|' in record.id and record.id.split('|')[1] in protein_cluster_to_keeps) or (record.id in protein_cluster_to_keeps)]

        # Do not create fasta file when 0 sequences were kept.
        if len(consensus_new_records) > 0:
            # Create output proteome file for OTU.
            consensus_fasta_file = os.path.join(reference_proteins_consensus_fasta_path, proteomes_tax_name+'.faa')
            SeqIO.write(consensus_new_records, consensus_fasta_file, 'fasta')
        else:
            logger.info('|EsMeCaTa|clustering| 0 protein clusters %s, no consensus fasta created (%d on %d).', proteomes_tax_name, index+1, nb_taxa_names)
        del consensus_new_records

        if remove_tmp:
            shutil.rmtree(mmseqs_tmp_cluster)

    # Compute number of protein clusters kept.
    stat_file = os.path.join(output_folder, 'stat_number_clustering.tsv')
    compute_stat_clustering(output_folder, stat_file)

    endtime = time.time()
    duration = endtime - starttime
    clustering_metadata['esmecata_clustering_duration'] = duration
    clustering_metadata_file = os.path.join(output_folder, 'esmecata_metadata_clustering.json')
    if os.path.exists(clustering_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_clustering' in metadata_file]
        clustering_metadata_file = os.path.join(output_folder, 'esmecata_metadata_clustering_{0}.json'.format(len(metadata_files)))
        with open(clustering_metadata_file, 'w') as ouput_file:
            json.dump(clustering_metadata, ouput_file, indent=4)
    else:
        with open(clustering_metadata_file, 'w') as ouput_file:
            json.dump(clustering_metadata, ouput_file, indent=4)

    logger.info('|EsMeCaTa|clustering| Clustering complete in {0}s.'.format(duration))



import csv
import os
import subprocess
import shutil

from esmecata.precomputed.create_database import create_database_from_esmecata_workflow_run
from esmecata.core.precomputed import precomputed_parse_affiliation
from esmecata.report.workflow_create_report import run_create_workflow_report

RESULTS = {
    'Cluster_1': {'proteomes': 2, 'protein_clusters': 603, 'GOs': 891, 'ECs': 313}
}

def test_workflow_online():
    subprocess.call(['esmecata', 'workflow_uniprot', '-i', 'buchnera_workflow.tsv', '-o', 'test_output', '-p', '1', '--minimal-nb-proteomes', '2', '--bioservices'])

    output_stat_file = os.path.join('test_output', 'stat_number_workflow.tsv')

    expected_results = {}
    with open(output_stat_file, 'r') as stat_file_read:
        csvreader = csv.reader(stat_file_read, delimiter='\t')
        next(csvreader)
        for line in csvreader:
            expected_results[line[0]] = {}
            expected_results[line[0]]['proteomes'] = int(line[5])
            expected_results[line[0]]['protein_clusters'] = int(line[7])
            expected_results[line[0]]['GOs'] = int(line[8])
            expected_results[line[0]]['ECs'] = int(line[9])

    for observation_name in expected_results:
        for data in expected_results[observation_name]:
            assert expected_results[observation_name][data] == RESULTS[observation_name][data]


def test_create_database_from_workflow_online():
    esmecata_result_folder = 'test_output'
    esmecata_database_output_folder = 'output_database'
    create_database_from_esmecata_workflow_run(esmecata_result_folder, esmecata_database_output_folder, '0.1.0')

    esmecata_database = os.path.join(esmecata_database_output_folder, 'esmecata_database.zip')
    output_folder = 'output_folder'
    precomputed_parse_affiliation('buchnera_workflow.tsv', esmecata_database, output_folder)

    report_output_folder = 'output_report_folder'
    run_create_workflow_report('buchnera_workflow.tsv', output_folder, report_output_folder)

    html_report_file = os.path.join(report_output_folder, 'esmecata_summary.html')
    # Check creation of HTML report file.
    assert os.path.exists(html_report_file)


def test_create_database_from_workflow_cli_online():
    esmecata_result_folder = 'test_output'
    esmecata_database_output_folder = 'output_database'
    subprocess.call(['esmecata_create_db', 'from_workflow', '-i', esmecata_result_folder, '-o', esmecata_database_output_folder, '--db-version', '0.1.0'])

    esmecata_database = os.path.join(esmecata_database_output_folder, 'esmecata_database.zip')
    output_folder = 'output_folder'
    precomputed_parse_affiliation('buchnera_workflow.tsv', esmecata_database, output_folder)

    report_output_folder = 'output_report_folder'
    run_create_workflow_report('buchnera_workflow.tsv', output_folder, report_output_folder)

    html_report_file = os.path.join(report_output_folder, 'esmecata_summary.html')
    # Check creation of HTML report file.
    assert os.path.exists(html_report_file)

    shutil.rmtree('test_output')


if __name__ == "__main__":
    test_workflow_online()



# Copyright (C) 2021-2025 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import argparse
import logging
import os
import sys
import time

from esmecata.precomputed.create_database import create_database_from_esmecata_run, merge_db_files
from esmecata.utils import is_valid_dir
from esmecata import __version__ as VERSION

MESSAGE = '''
Create database file from esmecata run.
'''
REQUIRES = '''
Requires: esmecata, pandas.
'''

logger = logging.getLogger()
logger.setLevel(logging.DEBUG)

def main():
    start_time = time.time()

    parser = argparse.ArgumentParser(
        'esmecata_create_db',
        description=MESSAGE + ' For specific help on each subcommand use: esmecata {cmd} --help',
        epilog=REQUIRES
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s ' + VERSION + '\n')

    parent_parser_i = argparse.ArgumentParser(add_help=False)
    parent_parser_i.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='EsMeCaTa output folder.',
        metavar='INPUT_FOLDER')

    parent_parser_iproteomes = argparse.ArgumentParser(add_help=False)
    parent_parser_iproteomes.add_argument(
        '-iproteomes',
        dest='input_proteomes',
        required=True,
        help='EsMeCaTa proteomes output folder.',
        metavar='INPUT_FOLDER')

    parent_parser_iclustering = argparse.ArgumentParser(add_help=False)
    parent_parser_iclustering.add_argument(
        '-iclustering',
        dest='input_clustering',
        required=True,
        help='EsMeCaTa clustering output folder.',
        metavar='INPUT_FOLDER')

    parent_parser_iannotation = argparse.ArgumentParser(add_help=False)
    parent_parser_iannotation.add_argument(
        '-iannotation',
        dest='input_annotation',
        required=True,
        help='EsMeCaTa annotation output folder.',
        metavar='INPUT_FOLDER')

    parent_parser_i_zip = argparse.ArgumentParser(add_help=False)
    parent_parser_i_zip.add_argument(
        '-i',
        '--input',
        dest='input',
        required=True,
        help='Multiple path of esmecata database zip files separated by ",".',
        metavar='INPUT_FILES')

    parent_parser_o = argparse.ArgumentParser(add_help=False)
    parent_parser_o.add_argument(
        '-o',
        '--output',
        dest='output',
        required=True,
        help='Output directory path.',
        metavar='OUPUT_DIR')

    parent_parser_database_version = argparse.ArgumentParser(add_help=False)
    parent_parser_database_version.add_argument(
        '--db-version',
        dest='database_version',
        help='Version number for EsMeCaTa precomputed database.',
        required=True,
        type=str)

    parent_parser_c = argparse.ArgumentParser(add_help=False)
    parent_parser_c.add_argument(
        '-c',
        '--core',
        dest='core',
        help='Number of CPU-cores for multiprocessing.',
        required=False,
        type=int,
        default=1)

    # subparsers
    subparsers = parser.add_subparsers(
        title='subcommands',
        description='valid subcommands:',
        dest='cmd')

    from_workflow = subparsers.add_parser(
        'from_workflow',
        help='Create database from esmecata workflow output.',
        parents=[
            parent_parser_i, parent_parser_o, parent_parser_database_version, parent_parser_c
            ],
        allow_abbrev=False)

    from_esmecata_runs = subparsers.add_parser(
        'from_runs',
        help='Create database from esmecata proteomes, clustering and annotation output folders.',
        parents=[
            parent_parser_iproteomes, parent_parser_iclustering, parent_parser_iannotation, parent_parser_o,
            parent_parser_database_version, parent_parser_c
            ],
        allow_abbrev=False)

    merge_db = subparsers.add_parser(
        'merge_db',
        help='Merge multiple zip files corresponding to EsMeCaTa databases.',
        parents=[
            parent_parser_i_zip, parent_parser_o
            ],
        allow_abbrev=False)

    args = parser.parse_args()

    # If no argument print the help.
    if len(sys.argv) == 1 or len(sys.argv) == 0:
        parser.print_help()
        sys.exit(1)

    is_valid_dir(args.output)

    # add logger in file
    formatter = logging.Formatter('%(message)s')
    log_file_path = os.path.join(args.output, f'esmecata_create_db_{args.cmd}.log')
    file_handler = logging.FileHandler(log_file_path, 'w+')
    file_handler.setLevel(logging.INFO)
    file_handler.setFormatter(formatter)
    logger.addHandler(file_handler)
    # set up the default console logger
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(logging.INFO)
    console_handler.setFormatter(formatter)
    logger.addHandler(console_handler)

    if args.cmd == 'from_workflow':
        esmecata_proteomes_folder = os.path.join(args.input, '0_proteomes')
        esmecata_clustering_folder = os.path.join(args.input, '1_clustering')
        esmecata_annotation_folder = os.path.join(args.input, '2_annotation')
        create_database_from_esmecata_run(esmecata_proteomes_folder, esmecata_clustering_folder, esmecata_annotation_folder, args.output, args.database_version, args.core)
    elif args.cmd == 'merge_db':
        if ',' in args.input:
            list_db_files = args.input.split(',')
        if ' ' in args.input:
            list_db_files = args.input.split(' ')
        merge_db_files(list_db_files, args.output)
    elif args.cmd == 'from_runs':
        create_database_from_esmecata_run(args.input_proteomes, args.input_clustering, args.input_annotation,
                                          args.output, args.database_version, args.core)
  
    logger.info("--- Total runtime %.2f seconds ---" % (time.time() - start_time))
    logger.warning(f'--- Logs written in {log_file_path} ---')


if __name__ == '__main__':
    main()



# Copyright (C) 2021-2025 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import argparse
import csv
import datetime
import logging
import os
import urllib.request
import sys
import time

from io import StringIO
from SPARQLWrapper import SPARQLWrapper, TSV
from socket import timeout

from esmecata import __version__ as esmecata_version

MIN_VAL = 0
MAX_VAL = 1
URLLIB_HEADERS = {'User-Agent': 'EsMeCaTa annotation v' + esmecata_version + ', request by urllib package v' + urllib.request.__version__}

logger = logging.getLogger(__name__)


def range_limited_float_type(arg):
    """Type function for argparse - a float within some predefined bounds

    Args:
        arg: argparse argument

    Returns:
        arg: argparse argument
    """
    try:
        f = float(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be a floating point number")
    if f < MIN_VAL or f > MAX_VAL:
        raise argparse.ArgumentTypeError("Argument must be < " + str(MAX_VAL) + " and > " + str(MIN_VAL))
    return f


def limited_integer_type(arg):
    """Type function for argparse - an integer

    Args:
        arg: argparse argument

    Returns:
        arg: argparse argument
    """
    try:
        f = int(arg)
    except ValueError:
        raise argparse.ArgumentTypeError("Must be an integer number")
    return f


def is_valid_path(filepath):
    """Return True if filepath is valid
    
    Args:
        filepath (str): path to file
    
    Returns:
        bool: True if path exists, False otherwise
    """
    if filepath and not os.access(filepath, os.W_OK):
        try:
            open(filepath, 'w').close()
            os.unlink(filepath)
            return True
        except OSError:
            return False
    else:  # path is accessible
        return True


def is_valid_file(filepath):
    """Return True if filepath exists

    Args:
        filepath (str): path to file

    Returns:
        bool: True if path exists, False otherwise
    """
    try:
        open(filepath, 'r').close()
        return True
    except OSError:
        return False


def is_valid_dir(dirpath):
    """Return True if directory exists or can be created (then create it)
    
    Args:
        dirpath (str): path of directory

    Returns:
        bool: True if dir exists, False otherwise
    """
    if not os.path.isdir(dirpath):
        try:
            os.makedirs(dirpath)
            return True
        except OSError:
            return False
    else:
        return True


def urllib_query(request, nb_retry=5):
    """Use urllib to query UniProt.

    Args:
        request (urllib.request.Request): request object
        nb_retry (int): number of retry to perform (default = 5)

    Returns:
        response (http.client.HTTPResponse): response returns by requests
    """
    passed = False

    if nb_retry == 0:
        sys.exit('5 retry attempts have been performed but were not successful, so esmecata has been stopped. You may try to relaunch it using the same command esmecata should resume.')
    try:
        response = urllib.request.urlopen(request)
        passed = True
    except timeout:
        logger.critical('Timeout occurs for query, try to relaunch query.')
        time.sleep(10)
        urllib_query(request, nb_retry-1)

    if passed is True:
        return response


def get_rest_uniprot_release(options):
    """Get the release version and date of Uniprot and Trembl and also the date of the query.

    Args:
        options (dict): dictionary with metadata on the esmecata run

    Returns:
        dict: metadata of Uniprot release
    """
    uniprot_releases = {}

    # Get Uniprot release version
    uniprot_urllib_request = urllib.request.Request('https://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/complete/reldate.txt',
                                                    headers=URLLIB_HEADERS)
    uniprot_response = urllib_query(uniprot_urllib_request)
    uniprot_lines = uniprot_response.readlines()
    uniprot_release_number = uniprot_lines[0].decode('utf-8').split(' ')[3].replace('\n','')
    swissprot_release_number = uniprot_lines[1].decode('utf-8').split(' ')[2].replace('\n','')
    swissprot_release_date = uniprot_lines[1].decode('utf-8').split(' ')[4].replace('\n','')
    trembl_release_number = uniprot_lines[2].decode('utf-8').split(' ')[2].replace('\n','')
    trembl_release_date = uniprot_lines[2].decode('utf-8').split(' ')[4].replace('\n','')

    date = datetime.datetime.now().strftime('%d-%B-%Y %H:%M:%S')

    uniprot_releases['esmecata_query_system'] = 'REST queries on Uniprot'
    uniprot_releases['uniprot_release'] = uniprot_release_number
    uniprot_releases['access_time'] = date
    uniprot_releases['swissprot_release_number'] = swissprot_release_number
    uniprot_releases['swissprot_release_date'] = swissprot_release_date
    uniprot_releases['trembl_release_number'] = trembl_release_number
    uniprot_releases['trembl_release_date'] = trembl_release_date
    uniprot_releases['tool_options'] = options

    return uniprot_releases


def get_sparql_uniprot_release(uniprot_sparql_endpoint, options):
    """Get the release version and date of Uniprot and Trembl and also the date of the query.

    Args:
        uniprot_sparql_endpoint (str): SPARQL endpoint to uniprot database
        options (dict): dictionary with metadata on the esmecata run

    Returns:
        dict: metadata of Uniprot release
    """
    uniprot_releases = {}

    uniprot_sparql_query = """SELECT ?version
    WHERE
    {{
        [] <http://www.w3.org/2002/07/owl#versionInfo> ?version .
    }}
    """
    csvreader = send_uniprot_sparql_query(uniprot_sparql_query, uniprot_sparql_endpoint)

    uniprot_release_number = [line[0] for line in csvreader][0]
    date = datetime.datetime.now().strftime('%d-%B-%Y %H:%M:%S')

    uniprot_releases['esmecata_query_system'] = 'SPARQL queries'
    uniprot_releases['esmecata_sparql_endpoint'] = uniprot_sparql_endpoint
    uniprot_releases['uniprot_release'] = uniprot_release_number
    uniprot_releases['access_time'] = date
    uniprot_releases['tool_options'] = options

    return uniprot_releases


def send_uniprot_sparql_query(sparql_query, uniprot_sparql_endpoint):
    """Get the release version and date of Uniprot and Trembl and also the date of the query.

    Args:
        sparql_query (str): string containing SPARQL query
        uniprot_sparql_endpoint (str): SPARQL endpoint to uniprot database

    Returns:
        dict: metadata of Uniprot release
    """
    sparql = SPARQLWrapper(uniprot_sparql_endpoint)
    sparql.agent = 'EsMeCaTa proteomes v' + esmecata_version + ', wrapper=' + sparql.agent

    sparql.setQuery(sparql_query)
    # Parse output.
    sparql.setReturnFormat(TSV)

    results = sparql.query().convert().decode('utf-8')
    if results != '':
        csvreader = csv.reader(StringIO(results), delimiter='\t')
        # Avoid header.
        next(csvreader)
    else:
        csvreader = []

    return csvreader



import pandas as pd
import os
import shutil
import subprocess

from esmecata.gseapy.gseapy_orsum import taxon_rank_annotation_enrichment


def test_taxon_rank_annotation_enrichment_selected():
    input_folder = 'annotation_output'
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = 'taxa_list.tsv'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, taxa_lists_file=taxa_list_file, orsum_minterm_size=4)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cli():
    input_folder = 'annotation_output'
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = 'taxa_list.tsv'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--taxa-list', taxa_list_file, '--orsumMinTermSize', '4'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cutoff():
    input_folder = 'annotation_output'
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = 'taxa_list.tsv'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, taxa_lists_file=taxa_list_file, orsum_minterm_size=4, selected_adjust_pvalue_cutoff=0.1)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_selected_cutoff_cli():
    input_folder = 'annotation_output'
    output_folder = 'output_folder'
    grouping = 'selected'
    taxa_list_file = 'taxa_list.tsv'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--taxa-list', taxa_list_file, '--orsumMinTermSize', '4', '--gseapyCutOff', '0.1'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_tax_rank():
    input_folder = 'annotation_output'
    output_folder = 'output_folder'
    grouping = 'tax_rank'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    taxon_rank_annotation_enrichment(input_folder, output_folder, grouping, orsum_minterm_size=4)

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)


def test_taxon_rank_annotation_enrichment_tax_rank_cli():
    input_folder = 'annotation_output'
    output_folder = 'output_folder'
    grouping = 'tax_rank'
    expected_terms = ['1.1.1.86', '2.8.4.1']
    if not os.path.exists(output_folder):
        os.mkdir(output_folder)
    subprocess.call(['esmecata_gseapy', 'gseapy_enrichr', '-f', input_folder, '-o', output_folder, '--grouping', grouping, '--orsumMinTermSize', '4'])

    result_file = os.path.join(output_folder, 'orsum_output_folder', 'filteredResult-Summary.tsv')
    df = pd.read_csv(result_file, sep='\t')
    representing_terms = df['Representing term id']

    assert sorted(representing_terms) == sorted(expected_terms)
    shutil.rmtree(output_folder)



# Copyright (C) 2023-2025 Arnaud Belcour - Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import os
import sys
import pandas as pd
import numpy as np
import datetime
import time
import json
import pronto
import logging
import subprocess
import urllib.request
import gseapy

from Bio.ExPASy import Enzyme
from matplotlib import __version__ as matplotlib_version
from esmecata.core.proteomes import get_taxon_obs_name
from esmecata import __version__ as esmecata_version
from Bio import __version__ as biopython_version

logger = logging.getLogger(__name__)


def extract_annotation(annotation_reference_folder):
    """ From annotation reference folder, creates a dict indicating the file associated with each annotation.

    Args:
        annotation_reference_folder (str): path to annotation reference folder
        dataset_annotation_file_path (str): path to output dataset annotation file

    Returns:
        annotation_sets (dict): annotation dict: annotation as key and taxon name as value
    """
    annotation_sets = {}
    for annotation_file in os.listdir(annotation_reference_folder):
        annotation_file_name = os.path.splitext(annotation_file)[0]
        annotation_file_path = os.path.join(annotation_reference_folder, annotation_file)

        with open(annotation_file_path, 'r') as open_annotation_input_file_path:
            csvreader = csv.DictReader(open_annotation_input_file_path, delimiter='\t')

            for line in csvreader:
                gos = line['GO'].split(',')
                ecs = line['EC'].split(',')

                if gos != ['']:
                    for go in gos:
                        if go not in annotation_sets:
                            annotation_sets[go] = [annotation_file_name]
                        else:
                            annotation_sets[go].append(annotation_file_name)
                if ecs != ['']:
                    for ec in ecs:
                        if ec not in annotation_sets:
                            annotation_sets[ec] = [annotation_file_name]
                        else:
                            annotation_sets[ec].append(annotation_file_name)

    return annotation_sets


def get_annot_name(enzyme_datfile, gobasic_file):
    """ Extract annotation names from enzyme.dat and go-baisc.obo.

    Args:
        enzyme_datfile (str): path to enzyme.Dat file
        gobasic_file (str): path to go-baisc.obo file

    Returns:
        enzyme_names (dict): EC as key and EC name as value
        go_names (dict): GO term ID as key and GO term name as value
    """
    enzyme_names = {}
    with open(enzyme_datfile) as infile:
        enzymes = Enzyme.parse(infile)
        for enzyme in enzymes:
            enzyme_id = enzyme['ID']
            enzyme_name = enzyme['DE']
            enzyme_names[enzyme_id] = enzyme_name

    go_ontology = pronto.Ontology(gobasic_file)
    go_names = {}
    for go_term in go_ontology:
        if 'GO:' in go_term:
            go = go_ontology[go_term]
            go_names[go_term] = go.name

    return enzyme_names, go_names


def run_orsum(annotation_file_gmt_file, enrichr_module_phylum_output, orsum_output_folder, orsum_minterm_size=None):
    """ Run orsum to reduce list of enriched annotations from gseapy.

    Args:
        annotation_file_gmt_file (str): GMT file containing annotations and their labels
        enrichr_module_phylum_output (str): path to output of gseapy containing enriched annotations for each taxon
        orsum_output_folder (str): path to output folder for orsum
        orsum_minterm_size (int): option minTermSize of orsum
    """
    input_files = [os.path.join(enrichr_module_phylum_output, file) for file in os.listdir(enrichr_module_phylum_output)]
    orsum_cmd = ['orsum.py', '--gmt', annotation_file_gmt_file, '--files', *input_files, '--outputFolder', orsum_output_folder]

    if orsum_minterm_size is not None:
        orsum_cmd.append('--minTermSize')
        orsum_cmd.append(str(orsum_minterm_size))

    logger.info('|EsMeCaTa|gseapy_enrichr| orsum command: {0}'.format(' '.join(orsum_cmd)))
    subprocess.call(orsum_cmd)


def get_orsum_version():
    """Returns version of orsum.

    Returns:
        orsum_version (str): version of orsum.
    """
    orsum_version = None
    response = subprocess.Popen(['orsum.py', '-v'], stdout=subprocess.PIPE, start_new_session=True, universal_newlines="")
    for orsum_line in response.stdout:
        orsum_line_decoded = orsum_line.decode('utf-8')
        orsum_version = orsum_line_decoded.strip('\n')

    if orsum_version is None:
        logger.critical('|EsMeCaTa|gseapy_enrichr| esmecata could not find the version of orsum.')
        logger.critical('|EsMeCaTa|gseapy_enrichr| It is possibly an issue with the installation of orsum (maybe it is not in the PATH). Or it can be due to a change in the output of orsum.py -v command.')
        sys.exit()

    return orsum_version


def extract_organisms_selected(taxa_lists_file):
    """ Extract observation names from input file (each column corresponds to a set of observation name to search for enrichment).

    Args:
        taxa_lists_file (str): path to taxa list file

    Returns:
        taxa_lists (dict): dictionary containing as key the column name of the file and as value the list of observation names
    """
    taxa_lists = {}
    df = pd.read_csv(taxa_lists_file, sep='\t')
    df = df.replace(np.nan, '')
    for col in df.columns:
        taxa_lists[col] = [observation_name for observation_name in df[col].tolist() if observation_name != '']
    return taxa_lists


def taxon_rank_annotation_enrichment(annotation_folder, output_folder, grouping="tax_rank",
                                     taxon_rank='phylum', taxa_lists_file=None,
                                     enzyme_data_file=None, go_basic_obo_file=None, orsum_minterm_size=None,
                                     selected_adjust_pvalue_cutoff=0.05):
    """ Run an enrichment analysis on taxon from annotation results of esmecata using gseapy.
    Then filter this list with orsum.

    Args:
        annotation_folder (str): path to esmecata annotation folder
        output_folder (str): path to output folder
        grouping (str): grouping factor, either "tax_rank" or "selected"
        taxon_rank (str): taxon rank to cluster the observation name together (by default, phylum) when selecting grouping "tax_rank"
        taxa_lists_file (str): path to manually created groups of observation names when selecting grouping "selected"
        enzyme_data_file (str): path to expasy enzyme.dat file, if not given, download it
        go_basic_obo_file (str): path to Gene Ontology go-basic.obo file, if not given, download it
        orsum_minterm_size (int): option minTermSize of orsum
        selected_adjust_pvalue_cutoff (float): adjust-Pval cutoff for gseapy enrichr, default: 0.05
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|gseapy_enrichr| Begin enrichment analysis.')

    if grouping == "tax_rank":
        if taxon_rank is None:
            logger.critical('|EsMeCaTa|gseapy_enrichr| You have to specify a taxon rank for this analysis with --taxon-rank')
            sys.exit()
        taxon_ranks = ['species', 'genus', 'family', 'order', 'class', 'phylum', 'kingdom', 'superkingdom']
        if taxon_rank != 'phylum':
            if taxon_rank not in taxon_ranks:
                logger.critical('|EsMeCaTa|gseapy_enrichr| Incorrect taxon given {0}, possible ranks are: {1}'.format(taxon_rank, ','.join(taxon_ranks)))
                sys.exit()
    elif grouping == "selected":
        if taxa_lists_file is None:
            logger.critical('|EsMeCaTa|gseapy_enrichr| You have to specify a taxa lists file for this analysis with --taxa-list')
            sys.exit()
    elif grouping is None:
        logger.critical('|EsMeCaTa|gseapy_enrichr| You have to choose a grouping factor either "tax_rank" or "selected".')
        sys.exit()

    # Get metadata associated with run.
    options = {'annotation_folder': annotation_folder, 'output_folder': output_folder, 'grouping': grouping,
               'taxon_rank': taxon_rank, 'taxa_lists_file': taxa_lists_file,
               'enzyme_data_file': enzyme_data_file, 'go_basic_obo_file': go_basic_obo_file, 'orsum_minterm_size': orsum_minterm_size}

    options['tool_dependencies'] = {}
    options['tool_dependencies']['python_package'] = {}
    options['tool_dependencies']['python_package']['Python_version'] = sys.version
    options['tool_dependencies']['python_package']['esmecata'] = esmecata_version
    options['tool_dependencies']['python_package']['pronto'] = pronto.__version__
    options['tool_dependencies']['python_package']['gseapy'] = gseapy.__version__
    options['tool_dependencies']['python_package']['orsum'] = get_orsum_version()
    options['tool_dependencies']['python_package']['pandas'] = pd.__version__
    options['tool_dependencies']['python_package']['biopython'] = biopython_version
    options['tool_dependencies']['python_package']['matplotlib'] = matplotlib_version

    esmecata_metadata = {}
    date = datetime.datetime.now().strftime('%d-%B-%Y %H:%M:%S')
    esmecata_metadata['access_time'] = date
    esmecata_metadata['tool_options'] = options

    # If no enzyme.Dat available download it.
    if enzyme_data_file is None:
        enzyme_data_file = os.path.join(output_folder, 'enzyme.dat')
        urllib.request.urlretrieve('https://ftp.expasy.org/databases/enzyme/enzyme.dat', enzyme_data_file)
    # If no go-basic.obo file available, download it.
    if go_basic_obo_file is None:
        go_basic_obo_file = os.path.join(output_folder, 'go-basic.obo')
        urllib.request.urlretrieve('http://purl.obolibrary.org/obo/go/go-basic.obo', go_basic_obo_file)

    enzyme_names, go_names = get_annot_name(enzyme_data_file, go_basic_obo_file)

    proteome_tax_id_file_path = os.path.join(annotation_folder, 'proteome_tax_id.tsv')

    if grouping == "tax_rank":
        taxa_name = get_taxon_obs_name(proteome_tax_id_file_path, taxon_rank)
    elif grouping == "selected":
        taxa_name = extract_organisms_selected(taxa_lists_file)

    annotation_reference_path = os.path.join(annotation_folder, 'annotation_reference')
    annotation_sets = extract_annotation(annotation_reference_path)

    # Create GMT file for orsum.
    # The GMT file contains each annotation, their labels and the observation name in which they were found.
    annotation_file_gmt_file = os.path.join(output_folder, 'annotation_file.gmt')
    with open(annotation_file_gmt_file, 'w') as open_output_file:
        csvwriter = csv.writer(open_output_file, delimiter='\t')
        for annotation in annotation_sets:
            annot_name = 'na'
            if annotation in enzyme_names:
                annot_name = annotation + ' ' + enzyme_names[annotation]
            if annotation in go_names:
                annot_name = annotation + ' ' + go_names[annotation]
            csvwriter.writerow([annotation, annot_name, *annotation_sets[annotation]])

    output_dir = os.path.join(output_folder, 'enrichr_module')
    orsum_input_folder = os.path.join(output_folder, 'orsum_input_folder')

    if not os.path.exists(output_dir):
        os.mkdir(output_dir)
    if not os.path.exists(orsum_input_folder):
        os.mkdir(orsum_input_folder)

    enriched_elements = {}
    for tax_name in taxa_name:
        organisms = taxa_name[tax_name]
        # Try to run gseapy enrichr, if no enriched results, continue.
        try:
            gseapy.enrichr(gene_list=organisms, gene_sets=annotation_sets, background=None,
                        outdir=os.path.join(output_dir, tax_name), cutoff=selected_adjust_pvalue_cutoff)
        except ValueError as error:
            logger.info('|EsMeCaTa|gseapy_enrichr| No enrichred functions with p-value cutoff < {0} for {1}.'.format(selected_adjust_pvalue_cutoff, tax_name))
            continue
        if os.path.exists(os.path.join(output_dir, tax_name, 'gs_ind_0.human.enrichr.reports.pdf')):
            # If enriched results, extract the ones with an adjusted p-value inferior to selected_adjust_pvalue_cutoff to output folder.
            df = pd.read_csv(os.path.join(output_dir, tax_name, 'gs_ind_0.human.enrichr.reports.txt'), sep='\t')
            df = df[df['Adjusted P-value'] < selected_adjust_pvalue_cutoff]
            enriched_elements[tax_name] = df.set_index('Term')['Adjusted P-value'].to_dict()

            df.sort_values('Adjusted P-value', inplace=True)
            df = df['Term']
            df.to_csv(os.path.join(orsum_input_folder, tax_name), sep='\t', index=False, header=False)

    all_elments = set([element for org in enriched_elements for element in enriched_elements[org]])

    enrich_matrix_file = os.path.join(output_folder, 'enrich_matrix.tsv')
    with open(enrich_matrix_file, 'w') as open_output_file:
        csvwriter = csv.writer(open_output_file, delimiter='\t')
        csvwriter.writerow(['Organism', *all_elments])
        for org in enriched_elements:
            csvwriter.writerow([org, *[enriched_elements[org][element] if element in enriched_elements[org] else 'NA' for element in all_elments]])

    # Check that there are files in orsum folder.
    orsum_input_folder_files = os.listdir(orsum_input_folder)
    if len(orsum_input_folder_files) == 0:
        logger.critical('|EsMeCaTa|gseapy_enrichr| No enriched files from gseapy enrichr as input for orsum. It seems that there are no enrichd terms found.')
        sys.exit(1)

    # Run orsum to filter list of enriched annotations.
    logger.info('|EsMeCaTa|gseapy_enrichr| Launch orsum visualisation.')
    orsum_output_folder = os.path.join(output_folder, 'orsum_output_folder')
    run_orsum(annotation_file_gmt_file, orsum_input_folder, orsum_output_folder, orsum_minterm_size)

    endtime = time.time()
    duration = endtime - starttime
    logger.info('|EsMeCaTa|gseapy_enrichr| Enrichment analysis took {0}.'.format(duration))

    esmecata_metadata['esmecata_gseapy_enrichr_duration'] = duration
    gseapy_enrichr_metadata_file = os.path.join(output_folder, 'esmecata_metadata_gseapy_enrichr.json')
    if os.path.exists(gseapy_enrichr_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_gseapy_enrichr' in metadata_file]
        gseapy_enrichr_metadata_file = os.path.join(output_folder, 'esmecata_metadata_gseapy_enrichr{0}.json'.format(len(metadata_files)))
        with open(gseapy_enrichr_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)
    else:
        with open(gseapy_enrichr_metadata_file, 'w') as ouput_file:
            json.dump(esmecata_metadata, ouput_file, indent=4)


# Copyright (C) 2021-2025 Arnaud Belcour - Inria, Univ Rennes, CNRS, IRISA Dyliss
# Univ. Grenoble Alpes, Inria, Microcosme
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>

import csv
import json
import logging
import os
import time

from esmecata.core.proteomes import retrieve_proteomes, compute_stat_proteomes
from esmecata.core.clustering import make_clustering, compute_stat_clustering
from esmecata.core.annotation import annotate_proteins, compute_stat_annotation
from esmecata.core.eggnog import annotate_with_eggnog

logger = logging.getLogger(__name__)


def compute_stat_workflow(proteomes_output_folder, clustering_output_folder, annotation_output_folder, stat_file=None):
    """Compute stat associated with the number of proteome for each taxonomic affiliations.

    Args:
        proteomes_output_folder (str): pathname to the result folder containing subfolder containing proteomes
        clustering_output_folder (str): pathname to the result folder containing mmseqs results
        annotation_output_folder (str): pathname to the annotation reference folder containing annotations for each cluster
        stat_file (str): pathname to the tsv stat file

    Returns:
        workflow_numbers (dict): dict containing observation names (as key) associated with proteomes, protein clusters, GO Terms and EC (as value)
    """
    workflow_numbers = {}

    proteome_numbers = compute_stat_proteomes(proteomes_output_folder)

    clustering_numbers = compute_stat_clustering(clustering_output_folder)

    annotation_reference_folder = os.path.join(annotation_output_folder, 'annotation_reference')
    annotation_numbers = compute_stat_annotation(annotation_reference_folder)

    all_observation_names = {*proteome_numbers.keys(), *clustering_numbers.keys(), *annotation_numbers.keys()}
    for observation_name in all_observation_names:
        if observation_name in proteome_numbers:
            nb_proteomes, tax_name, lowest_tax_rank, esmecata_name, esmecata_rank, reference_status = proteome_numbers[observation_name]
        else:
            nb_proteomes = 'NA'
        if observation_name in clustering_numbers:
            nb_shared_proteins = len(clustering_numbers[observation_name][1])
        else:
            nb_shared_proteins = 'NA'
        if observation_name in annotation_numbers:
            nb_gos = annotation_numbers[observation_name][0]
            nb_ecs = annotation_numbers[observation_name][1]
        else:
            nb_gos = 'NA'
            nb_ecs = 'NA'
        workflow_numbers[observation_name] = [nb_proteomes, tax_name, lowest_tax_rank, esmecata_name, esmecata_rank, reference_status, nb_shared_proteins, nb_gos, nb_ecs]

    if stat_file:
        with open(stat_file, 'w') as stat_file_open:
            csvwriter = csv.writer(stat_file_open, delimiter='\t')
            csvwriter.writerow(['observation_name', 'Input_taxon_name', 'Input_rank', 'EsMeCaTa_chosen_name', 'EsMeCaTa_chosen_rank','Number_proteomes', 'Only_reference_proteome', 'Proteins_clusters_0_5' , 'Number_go_terms', 'Number_ecs'])
            for observation_name in workflow_numbers:
                nb_proteomes, tax_name, lowest_tax_rank, esmecata_name, esmecata_rank, reference_status = workflow_numbers[observation_name][:6]
                nb_shared_proteins = workflow_numbers[observation_name][6]
                nb_gos = workflow_numbers[observation_name][7]
                nb_ecs = workflow_numbers[observation_name][8]
                csvwriter.writerow([observation_name, tax_name, lowest_tax_rank, esmecata_name, esmecata_rank, nb_proteomes, reference_status, nb_shared_proteins, nb_gos, nb_ecs])

    return workflow_numbers


def perform_workflow(input_file, output_folder, busco_percentage_keep=80, ignore_taxadb_update=None,
                        all_proteomes=None, uniprot_sparql_endpoint=None, remove_tmp=None,
                        limit_maximal_number_proteomes=99, rank_limit=None,
                        nb_core=1, clust_threshold=1, mmseqs_options=None,
                        linclust=None, propagate_annotation=None, uniref_annotation=None,
                        expression_annotation=None, minimal_number_proteomes=1, annotation_files=None,
                        update_affiliations=None, option_bioservices=None):
    """From the proteomes found by esmecata proteomes, create protein cluster for each taxonomic affiliations.

    Args:
        input_file (str): pathname to the tsv input file containing taxonomic affiliations
        output_folder (str): pathname to the output folder
        busco_percentage_keep (float): BUSCO score to filter proteomes (proteomes selected will have a higher BUSCO score than this threshold)
        ignore_taxadb_update (bool): option to ignore ete3 taxa database update
        all_proteomes (bool): Option to select all the proteomes (and not only preferentially reference proteomes)
        uniprot_sparql_endpoint (str): uniprot SPARQL endpoint to query (by default query Uniprot SPARQL endpoint)
        remove_tmp (bool): remove the tmp files
        limit_maximal_number_proteomes (int): int threshold after which a subsampling will be performed on the data
        rank_limit (str): rank limit to filter the affiliations (keep this rank and all inferior ranks)
        nb_core (int): number of CPU-cores to be used by mmseqs
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        mmseqs_options (str): use alternative mmseqs option
        linclust (bool): use linclust
        propagate_annotation (float): float between 0 and 1. It is the ratio of proteins in the cluster that should have the annotation to keep this annotation.
        uniref_annotation (bool): option to use uniref annotation to add annotation.
        expression_annotation (bool): option to add expression annotation from UniProt.
        minimal_number_proteomes (int): minimal number of proteomes required to be associated with a taxon for the taxon to be kept.
        annotation_files (str): pathnames to UniProt dat files.
        update_affiliations (str): option to update taxonomic affiliations.
        option_bioservices (bool): use bioservices instead of manual queries.
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|workflow| Begin workflow.')
    workflow_metadata = {}

    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    proteomes_output_folder = os.path.join(output_folder, '0_proteomes')
    retrieve_proteomes(input_file, proteomes_output_folder, busco_percentage_keep,
                        ignore_taxadb_update, all_proteomes, uniprot_sparql_endpoint,
                        limit_maximal_number_proteomes, rank_limit, minimal_number_proteomes,
                        update_affiliations, option_bioservices)

    clustering_output_folder = os.path.join(output_folder, '1_clustering')
    make_clustering(proteomes_output_folder, clustering_output_folder, nb_core, clust_threshold, mmseqs_options, linclust, remove_tmp)

    annotation_output_folder = os.path.join(output_folder, '2_annotation')
    annotate_proteins(clustering_output_folder, annotation_output_folder, uniprot_sparql_endpoint, propagate_annotation,
                      uniref_annotation, expression_annotation, annotation_files, option_bioservices)

    stat_file = os.path.join(output_folder, 'stat_number_workflow.tsv')
    compute_stat_workflow(proteomes_output_folder, clustering_output_folder, annotation_output_folder, stat_file)

    proteomes_metadata_file = os.path.join(proteomes_output_folder, 'esmecata_metadata_proteomes.json')
    with open(proteomes_metadata_file, 'r') as json_idata:
        proteomes_metadata = json.load(json_idata)
    clustering_metadata_file = os.path.join(clustering_output_folder, 'esmecata_metadata_clustering.json')
    with open(clustering_metadata_file, 'r') as json_idata:
        clustering_metadata = json.load(json_idata)
    annotation_metadata_file = os.path.join(annotation_output_folder, 'esmecata_metadata_annotation.json')
    with open(annotation_metadata_file, 'r') as json_idata:
        annotation_metadata = json.load(json_idata)

    workflow_metadata['proteomes_metadata'] = proteomes_metadata
    workflow_metadata['clustering_metadata'] = clustering_metadata
    workflow_metadata['annotation_metadata'] = annotation_metadata

    endtime = time.time()
    duration = endtime - starttime
    workflow_metadata['esmecata_workflow_duration'] = duration
    workflow_metadata_file = os.path.join(output_folder, 'esmecata_metadata_workflow.json')
    if os.path.exists(workflow_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_workflow' in metadata_file]
        workflow_metadata_file = os.path.join(output_folder, 'esmecata_metadata_workflow_{0}.json'.format(len(metadata_files)))
        with open(workflow_metadata_file, 'w') as ouput_file:
            json.dump(workflow_metadata, ouput_file, indent=4)
    else:
        with open(workflow_metadata_file, 'w') as ouput_file:
            json.dump(workflow_metadata, ouput_file, indent=4)

    logger.info('|EsMeCaTa|workflow| Workflow step complete in {0}s.'.format(duration))



def perform_workflow_eggnog(input_file, output_folder, eggnog_database_path, busco_percentage_keep=80,
                            ignore_taxadb_update=None, all_proteomes=None, uniprot_sparql_endpoint=None,
                            remove_tmp=None, limit_maximal_number_proteomes=99, rank_limit=None,
                            nb_core=1, clust_threshold=0.5, mmseqs_options=None,
                            linclust=None, minimal_number_proteomes=5, update_affiliations=None,
                            option_bioservices=None, eggnog_tmp_dir=None, no_dbmem=False):
    """From the proteomes found by esmecata proteomes, create protein cluster for each taxonomic affiliations.

    Args:
        input_file (str): pathname to the tsv input file containing taxonomic affiliations
        output_folder (str): pathname to the output folder
        eggnog_database_path (str): pathname to eggnog database folder
        busco_percentage_keep (float): BUSCO score to filter proteomes (proteomes selected will have a higher BUSCO score than this threshold)
        ignore_taxadb_update (bool): option to ignore ete3 taxa database update
        all_proteomes (bool): Option to select all the proteomes (and not only preferentially reference proteomes)
        uniprot_sparql_endpoint (str): uniprot SPARQL endpoint to query (by default query Uniprot SPARQL endpoint)
        remove_tmp (bool): remove the tmp files
        limit_maximal_number_proteomes (int): int threshold after which a subsampling will be performed on the data
        rank_limit (str): rank limit to filter the affiliations (keep this rank and all inferior ranks)
        nb_core (int): number of CPU-cores to be used by mmseqs and eggnog-mapper
        clust_threshold (float): threshold to select protein cluster according to the representation of protein proteome in the cluster
        mmseqs_options (str): use alternative mmseqs option
        linclust (bool): use linclust
        minimal_number_proteomes (int): minimal number of proteomes required to be associated with a taxon for the taxon to be kept.
        update_affiliations (str): option to update taxonomic affiliations.
        option_bioservices (bool): use bioservices instead of manual queries.
        eggnog_tmp_dir (str): pathname to eggnog-mapper temporary folder.
        no_dbmem (bool): Boolean to choose to not load eggnog database in memory.
    """
    starttime = time.time()
    logger.info('|EsMeCaTa|workflow| Begin workflow.')
    workflow_metadata = {}

    if not os.path.exists(output_folder):
        os.mkdir(output_folder)

    proteomes_output_folder = os.path.join(output_folder, '0_proteomes')
    retrieve_proteomes(input_file, proteomes_output_folder, busco_percentage_keep,
                        ignore_taxadb_update, all_proteomes, uniprot_sparql_endpoint,
                        limit_maximal_number_proteomes, rank_limit, minimal_number_proteomes,
                        update_affiliations, option_bioservices)

    clustering_output_folder = os.path.join(output_folder, '1_clustering')
    make_clustering(proteomes_output_folder, clustering_output_folder, nb_core, clust_threshold, mmseqs_options, linclust, remove_tmp)

    annotation_output_folder = os.path.join(output_folder, '2_annotation')
    annotate_with_eggnog(clustering_output_folder, annotation_output_folder, eggnog_database_path, nb_core, eggnog_tmp_dir, no_dbmem)

    stat_file = os.path.join(output_folder, 'stat_number_workflow.tsv')
    compute_stat_workflow(proteomes_output_folder, clustering_output_folder, annotation_output_folder, stat_file)

    proteomes_metadata_file = os.path.join(proteomes_output_folder, 'esmecata_metadata_proteomes.json')
    with open(proteomes_metadata_file, 'r') as json_idata:
        proteomes_metadata = json.load(json_idata)
    clustering_metadata_file = os.path.join(clustering_output_folder, 'esmecata_metadata_clustering.json')
    with open(clustering_metadata_file, 'r') as json_idata:
        clustering_metadata = json.load(json_idata)
    annotation_metadata_file = os.path.join(annotation_output_folder, 'esmecata_metadata_annotation.json')
    with open(annotation_metadata_file, 'r') as json_idata:
        annotation_metadata = json.load(json_idata)

    workflow_metadata['proteomes_metadata'] = proteomes_metadata
    workflow_metadata['clustering_metadata'] = clustering_metadata
    workflow_metadata['annotation_metadata'] = annotation_metadata

    endtime = time.time()
    duration = endtime - starttime
    workflow_metadata['esmecata_workflow_duration'] = duration
    workflow_metadata_file = os.path.join(output_folder, 'esmecata_metadata_workflow.json')
    if os.path.exists(workflow_metadata_file):
        metadata_files = [metadata_file for metadata_file in os.listdir(output_folder) if 'esmecata_metadata_workflow' in metadata_file]
        workflow_metadata_file = os.path.join(output_folder, 'esmecata_metadata_workflow_{0}.json'.format(len(metadata_files)))
        with open(workflow_metadata_file, 'w') as ouput_file:
            json.dump(workflow_metadata, ouput_file, indent=4)
    else:
        with open(workflow_metadata_file, 'w') as ouput_file:
            json.dump(workflow_metadata, ouput_file, indent=4)

    logger.info('|EsMeCaTa|workflow| Workflow step complete in {0}s.'.format(duration))
